#ifndef DISKCACHE_API__HOT_DIRECTORY_HH
#define DISKCACHE_API__HOT_DIRECTORY_HH

#include "general/SharedPtr.hh"
#include "dirinfo.hh"

namespace diskCache
{
  //---------------------------------------------------------------------
  /// \brief Information about a directory that is being actively watched.
  ///
  /// This maintains a collection of hot directories.
  /// Hot directories are scanned outside of the conventional deep scan
  /// of the MountPoint class so as to reduce discovery latency.
  //---------------------------------------------------------------------
  class HotDirectory
  {
  public:
    //-------------------------------------------------------------------
    /// \brief Storage type for directory information.
    ///
    /// Having the directory information stored in a shared pointer
    /// container increases thread safety and diminishes the needs
    /// for locks.
    //-------------------------------------------------------------------
    typedef General::WeakPtr< DirectoryInfo >	weak_dir_info_type;
    //-------------------------------------------------------------------
    /// \brief Collection of patterns used to exclude directories
    ///
    /// This information needs to be preserved for doing a scan
    /// of a directory.
    /// It should be treated as an opaque object.
    //-------------------------------------------------------------------
    typedef DirectoryInfo::dir_container_type	dir_container_type;
    
    //-------------------------------------------------------------------
    /// \brief Lower bound of relative age in seconds to be considered.
    ///
    /// \return
    ///     The current lower bound.
    //-------------------------------------------------------------------
    static mtime_type HotLowerBound( );

    //-------------------------------------------------------------------
    /// \brief Set lower bound of relative age in seconds to be considered.
    ///
    /// \param[in] Value
    ///     The new lower bound.
    ///     A value of 0 (zero) disables hot directories.
    //-------------------------------------------------------------------
    static void HotLowerBound( mtime_type Value );

    //-------------------------------------------------------------------
    /// \brief Report if the path is registered as hot
    ///
    /// \param[in] Path
    ///     The path to check.
    //-------------------------------------------------------------------
    static bool IsRegistered( const std::string& Path );

    //-------------------------------------------------------------------
    /// \brief Add a directory to the list of hot directories.
    ///
    /// \param[in] Entry
    ///     The Entry to add to the list of hot directories if and
    ///     only if it meets all hot directory criteria.
    /// \param[in] ExcludeDirs
    ///     A list of known directory names that should not be scanned.
    /// \param[in] Log
    ///     Set to true if the directory should log a message if
    ///     Entry is added to the hot list.
    //-------------------------------------------------------------------
    static void RegisterIfHot( weak_dir_info_type Entry,
			       const dir_container_type& ExcludeDirs,
			       bool Log );

    //-------------------------------------------------------------------
    /// \brief Establish the frequency that hot directories are scanned.
    ///
    /// \param[in] Value
    ///     The new scan interval.
    //-------------------------------------------------------------------
    static void ScanInterval( mtime_type Value );

    //-------------------------------------------------------------------
    /// \brief The frequency that hot directories are scanned.
    ///
    /// \return
    ///     The scan interval.
    //-------------------------------------------------------------------
    static mtime_type ScanInterval( );

    //-------------------------------------------------------------------
    /// \brief Remove a directory from the list of hot directories.
    ///
    /// \param[in]
    ///     The Entry to be removed from the list of hot directories.
    //-------------------------------------------------------------------
    static void Unregister( weak_dir_info_type Entry );

  private:
    //-------------------------------------------------------------------
    /// \brief Storage for lower bound of relative age in seconds.
    //-------------------------------------------------------------------
    static mtime_type m_lower_bound;

    static mtime_type m_scan_interval;

  };

  //---------------------------------------------------------------------
  /// Return the time representing the number of seconds to sleep
  /// between scans of the set of hot directories.
  //---------------------------------------------------------------------
  inline mtime_type HotDirectory::
  ScanInterval( )
  {
    return m_scan_interval;
  }

}

#endif /* DISKCACHE_API__HOT_DIRECTORY_HH */
