// -*- mode: c++ basic-offset: 2; indent-tabs-mode: nil; -*-

#include "config.h"

#include <cassert>

#include <iostream>

#include <algorithm>
#include <functional>
#include <list>
#include <map>
#include <memory>
#include <set>

#include "general/regex.hh"
#include "general/regexmatch.hh"
#include "general/SharedPtr.hh"
#include "general/unordered_map.hh"

#include "genericAPI/Logging.hh"

#include "diskcache.hh"
#include "dirinfo.hh"
#include "diskcache.hh"
#include "FilenameHandler.hh"
#include "mountpt.hh"
#include "Query.hh"
#include "QueryResults.hh"
#include "QueryResultsDump.hh"
#include "QueryResultsFilenames.hh"
#include "QueryResultsIntervals.hh"
#include "QueryResultsMountPointStats.hh"
#include "QueryResultsTimeRange.hh"
#include "SDGTxFilenameHandler.hh"
#include "timehash.hh"

using std::runtime_error;
using std::string;
using std::unary_function;

using GenericAPI::queueLogEntry;
using GenericAPI::LogEntryGroup_type;

#define SDGTx_VERBOSE_DEBUGGING 1

typedef diskCache::CacheFile CacheFile;
typedef diskCache::DirectoryInfo DirectoryInfo;
typedef diskCache::FilenameHandler::Hash Hash;
typedef diskCache::LogMessage LogMessage;
typedef diskCache::MountPointHash MountPointHash;
typedef diskCache::MountPointCache MountPointCache;
typedef diskCache::Query Query;
typedef diskCache::QueryResults QueryResults;
typedef diskCache::TimeHash TimeHash;
typedef diskCache::dump_format_type dump_format_type;
typedef diskCache::mtime_type mtime_type;

typedef diskCache::TimeHash::KnownIntervals conflict_type;

//-----------------------------------------------------------------------
// forward declarations
//-----------------------------------------------------------------------

namespace {
  class lHash;
}

namespace std
{
  std::ostream& operator<<( std::ostream& Stream, const lHash& Data );
}

//-----------------------------------------------------------------------
/// \brief Default pattern used for SDGTx filename matching
//-----------------------------------------------------------------------
static const char* sdgt_gwf_regex_pattern
=
  "^([^\\-]+)\\-([^\\-]+)\\-([[:digit:]]+)\\-([[:digit:]]+)([.]gwf)$"
  ;

//-----------------------------------------------------------------------
/// \brief Number of extected matches
//-----------------------------------------------------------------------
const static size_t SDGTx_MATCH_SIZE = 6;

//-----------------------------------------------------------------------
/// \brief Regular expression used for testing the pattern
///
/// \todo
///     Use of this regular expression needs to be protected by
///     locks.
/// \todo
///     The Regex class needs to be extened to support a reset option
///     so the regular expression may be modified if the list of
///     file extensions is modified.
//-----------------------------------------------------------------------
static Regex SDGTx_RegEx( sdgt_gwf_regex_pattern, REG_EXTENDED );

enum {
  SDGTx_S_OFFSET = 1,
  SDGTx_D_OFFSET,
  SDGTx_G_OFFSET,
  SDGTx_T_OFFSET,
  SDGTx_x_OFFSET
};

namespace {
  using diskCache::QueryResultsTimeRange;

  typedef std::vector< diskCache::TimeHash::KnownIntervals > known_interval_container_type;
  typedef std::map< INT_4U, std::list< INT_4U > > time_hash_data_src_type;

  inline INT_4U ifloor( INT_4U Target, INT_4U Start, INT_4U DT )
  {
    if ( Target < Start )
    {
      return Start;
    }
    return Start + ( ( ( Target - Start ) / DT ) * DT );
  }

  inline INT_4U iceil( INT_4U Target, INT_4U Stop, INT_4U DT )
  {
    return Stop - ( ( ( Stop - Target ) / DT ) * DT );
  }

  class FilenameData
    : public QueryResultsTimeRange::RangeData
  {
  public:
    const std::string                           s_directory;
    const std::string                           s_ifo_type;
    const std::string                           s_extension;
    INT_4U                                      s_mount_point;

    FilenameData( const std::string& Directory,
                  const std::string& IFOType,
                  const std::string& Extension,
                  Query::gps_time_type EndTime,
                  Query::gps_time_type DeltaT,
                  INT_4U MountPointSize );

    virtual ~FilenameData( );

    static void Reconstitute( Query::gps_time_type Start,
                              const QueryResultsTimeRange::RangeData* Data,
                              QueryResultsTimeRange& Results );
  };

  inline FilenameData::
  FilenameData( const std::string& Directory,
                const std::string& IFOType,
                const std::string& Extension,
                Query::gps_time_type EndTime,
                Query::gps_time_type DeltaT,
                INT_4U MountPointSize )
    : QueryResultsTimeRange::RangeData( EndTime, DeltaT ),
      s_directory( Directory ),
      s_ifo_type( IFOType ),
      s_extension( Extension ),
      s_mount_point( MountPointSize )
  {
  }

  FilenameData::
  ~FilenameData( )
  {
  }

  void FilenameData::
  Reconstitute( Query::gps_time_type StartTime,
                const QueryResultsTimeRange::RangeData* Data,
                QueryResultsTimeRange& Results )
  {
    const FilenameData* d = reinterpret_cast< const FilenameData* >( Data );
    for ( Query::gps_time_type
            cur_gps = StartTime,
            last_gps = d->EndTime( ),
            inc_gps = d->DeltaT( );
          cur_gps < last_gps;
          cur_gps += inc_gps )
    {
      std::ostringstream        filename;

      filename << d->s_directory << "/" << d->s_ifo_type
               << cur_gps << "-" << inc_gps << d->s_extension
        ;
      Results.Append( filename.str( ) );
    }
  }


#if ENABLE_MOUNT_POINT_CACHE
  //---------------------------------------------------------------------
  // Forward declaration of classes
  //---------------------------------------------------------------------
  class lMountPointCache;
#endif /* ENABLE_MOUNT_POINT_CACHE */

  //---------------------------------------------------------------------
  // local data
  //---------------------------------------------------------------------
  inline const char* SEPERATOR( )
  {
    static const char* value = "-";

    return value;
 }
  static const std::string GLOB_ALL( "*" );

  //---------------------------------------------------------------------
  //---------------------------------------------------------------------
  class qr_intervals
    : public diskCache::QueryResults
  {
  public:
    qr_intervals( format_type Format,
                  type_type Type,
                  const DirectoryInfo* ExcludeDirectory,
                  known_interval_container_type& KnownIntervals );

    void Append( const diskCache::TimeHash::KnownIntervals& Data );

    const DirectoryInfo*                m_exclude_directory;

  private:
    known_interval_container_type&      m_known_intervals;
  };

  qr_intervals::
  qr_intervals( format_type Format,
                type_type Type,
                const DirectoryInfo* ExcludeDirectory,
                known_interval_container_type& KnownIntervals )
    : QueryResults( Format, Type ),
      m_exclude_directory( ExcludeDirectory ),
      m_known_intervals( m_known_intervals )
  {
  }

  inline void qr_intervals::
  Append( const diskCache::TimeHash::KnownIntervals& Data )
  {
    m_known_intervals.push_back( Data );
  }

  //---------------------------------------------------------------------
  //---------------------------------------------------------------------

  //: Predicate used to check if a pair of ifo and type 
  //: satisfy a searched criteria.
  class IfoTypeMatch
    : public unary_function< string, bool >
  {
  public:
      
    IfoTypeMatch( const string& ifo, const string& type );
   
    bool operator()( const string& ifo_type ) const;
   
   
  private:

    static const string mAll;
   
    //: A requested ifo for frame data
    string mIfo; 
      
    bool mIfoAll;

    //: A requested type for frame data
    string mType;
   
    bool mTypeAll;
    //: A flag to indicate if explicit ifo and type are requested
    //: (set to FALSE if one of "ifo" or "type" is "all" )
    bool mExplicit;
   
    //: Formated "IFO-Type-" string (used only if mExplicit is TRUE)
    string mIfoType;
  };
   
   
  //: Static data initialization
  const string IfoTypeMatch::mAll( "all-" );
   
   
  //: Constructor
  //
  //!param: const string& ifo - Searched IFO for frame data (could be "all").
  //!param: const string& type - Searched Type for frame data (could be "all").
  //
  IfoTypeMatch::
  IfoTypeMatch( const string& ifo, const string& type )
    : mIfo( ifo ),
      mIfoAll( false ),
      mType( type ),
      mTypeAll( false ),
      mExplicit( false )
  {
    // Include SEPERATOR to guarantee exact ifo and type match:
    if ( mType[ mType.length( ) - 1] != *(SEPERATOR( ) ) )
    {
      mType += SEPERATOR( );
    }
    mTypeAll = ( mType == mAll );

    if ( ( mIfo.length( ) == 2 ) &&
         ( mIfo[ 1 ] == *(SEPERATOR( ) ) ) )
    {
      mIfo.erase( 1 );
    }
    if ( mIfo.length( ) > 1 )
    {
      if ( mIfo[ mIfo.length( ) - 1] != *(SEPERATOR( ) ) )
      {
        mIfo += SEPERATOR( );
      }
      mIfoAll = ( mIfo == mAll );

      if( !mIfoAll && !mTypeAll )
      {
        mExplicit = true;
   
        // Format key used in hash lookups
        mIfoType = mIfo;
        mIfoType += mType;
      }   
    }
  }
   
   
  //: Compare operator
  //
  //!param: const string& ifo_type - "IFO-Type-" formated string..
  //
  //!return: bool - True if key matches searched ifo and type, false otherwise.
  //
  bool IfoTypeMatch::
  operator( )( const string& ifo_type ) const
  {
    static const CHAR* const method_name( "IfoTypeMatch::operator(): " );
   
    if( mExplicit )
    {
      return ( ifo_type == mIfoType );
    }
   

    string::size_type dash_pos( std::string::npos );

    // Evaluate "ifo" part of the key
    if( mIfoAll == false )
    {
      // Compare ifo part of key if "ifo" is not a single character ===>
      // "type" is set to "all-"
      if( mIfo.size( ) == 1 )
      {
        // Check if "ifo" part of the key contains specified "ifo"
        // character
   
        // Locate first name delimiter: between "IFO" and "Type"
        dash_pos = ifo_type.find( SEPERATOR( ) );
   
        // Sanity check: key should always be properly formatted, if not -
        // something is really wrong!!!
        if( dash_pos == string::npos )
        {
          string msg( method_name );
          msg += "Could not extract Type information from hash map key \'";
          msg += ifo_type;
          msg += "\'";
   
          throw runtime_error( msg.c_str() );
        } 
   
   
        // Check if ifo part of the key contains mIfo character
        string::size_type ifo_pos = ifo_type.find( mIfo );
   
        if( ( ifo_pos == string::npos )
            || ( ifo_pos >= dash_pos ) )
        {
          return false;
        }
      }
      // Specified "ifo" is not a single character   
      else if( ifo_type.compare( 0, mIfo.size( ), mIfo ) != 0 )
      {
        return false;
      }
    }


    // Evaluate "type" part of the key
    if( mTypeAll == false )
    {
      if ( dash_pos == std::string::npos )
      {
        dash_pos = ifo_type.find( SEPERATOR( ) );
      }
   
      // Sanity check: key should always be properly formatted, if not -
      // something is really wrong!!!
      if( dash_pos == string::npos )
      {
        string msg( method_name );
        msg += "Could not extract Type information from hash map key \'";
        msg += ifo_type;
        msg += "\'";
   
        throw runtime_error( msg.c_str() );
      }
   
   
      return ( ifo_type.compare( ++dash_pos, mType.size(), mType ) == 0 );
    }
   

    // Method would have already returned "false" if neither "ifo" nor "type"
    // matched requested query
    return true;
  }

  //=====================================================================
  /// \brief Index for SDGT.x file patterns
  //=====================================================================
  class lHash
    : public Hash
  {
  public:
    typedef diskCache::FilenameHandler::dir_info_type dir_info_type;
    typedef diskCache::FilenameHandler::IsFileData::conflict_container_type
    conflict_container_type;
    lHash( dir_info_type Value );

    virtual ~lHash();

    //-----------------------------------------------------------
    /// \brief Add an entry to the has
    ///
    /// \param[in] IFOType
    ///    The site identifier followed by the '-' character
    ///    followed by the description text followed by a
    ///    trailing '-' character.
    /// \param[in] GPSStart
    ///    The GPS start time in integral seconds.
    /// \param[in] DT
    ///    The number of whole seconds covered by the file.
    /// \param[in] Extension
    ///    The file extension associated with the file.
    //-----------------------------------------------------------
    void Add( const std::string& IFOType,
              INT_4U GPSStart,
              INT_4U DT,
              const std::string& Extension );

    INT_4U Add( const lHash& Source,
                conflict_container_type& Conflicts );

    virtual void Read( CacheFile& Stream );

    virtual std::ostream& Write( std::ostream& Stream ) const;

    virtual void Write( CacheFile& Stream ) const;

    //-----------------------------------------------------------
    /// \brief Determine if a file is currently part of the hash
    ///
    ///
    /// \param[in] IFOType
    ///    The site identifier followed by the '-' character
    ///    followed by the description text followed by a
    ///    trailing '-' character.
    /// \param[in] GPSStart
    ///    The GPS start time in integral seconds.
    /// \param[in] DT
    ///    The number of whole seconds covered by the file.
    /// \param[in] Extension
    ///    The file extension associated with the file.
    /// 
    /// \return
    ///     The value of true is returned if the file
    ///     with the above characteristics is currently
    ///     covered by the hash.
    //-----------------------------------------------------------
    bool Find( const std::string& IFOType,
               INT_4U GPSStart,
               INT_4U DT,
               const std::string& Extension ) const;

    virtual void Find( const Query& SearchParams,
                       QueryResults& Results ) const;

    virtual size_t FileCount( ) const;

    //-----------------------------------------------------------
    /// \brief Determine if a file is currently part of the hash
    ///
    ///
    /// \param[in] IFOType
    ///    The site identifier followed by the '-' character
    ///    followed by the description text followed by a
    ///    trailing '-' character.
    /// \param[in] GPSStart
    ///     The inclusive start time of the interval to search
    /// \param[in] GPSEnd
    ///     The exclusive end time of the interval to search
    /// \param[in] Extension
    ///    The file extension associated with the file.
    /// 
    /// \return
    ///      The value of true is returned if the file
    ///      with the above characteristics is currently
    ///      covered by the hash.
    //-----------------------------------------------------------
    bool Overlapping( const std::string& IFOType,
                      INT_4U GPSStart,
                      INT_4U GPSEnd,
                      const std::string& Extension ) const;

  private:
#if WORKING
    friend void lMountPointCache::
    Register( General::SharedPtr< diskCache::FilenameHandler::Hash >
              DirectoryInfoHash,
              int CheckForConflicts );
#else  /* WORKING */
    friend class lMountPointCache;
#endif /* WORKING */
    //-------------------------------------------------------------------
    /// \brief time index type
    //-------------------------------------------------------------------
    typedef General::SharedPtr< TimeHash > time_data_type;
    //-------------------------------------------------------------------
    /// \brief IFO-Type- type
    //-------------------------------------------------------------------
    typedef std::string	site_desc_type;
    //-------------------------------------------------------------------
    /// \brief Filename extension type
    //-------------------------------------------------------------------
    typedef std::string	ext_type;
    //-------------------------------------------------------------------
    /// \brief Ordered container of site/desc info and time hash
    //-------------------------------------------------------------------
    typedef std::map< site_desc_type, time_data_type >  file_base_container_type;
      
    //-------------------------------------------------------------------
    /// \brief Unordered container of file extension and file base names
    //-------------------------------------------------------------------
    typedef General::unordered_map< ext_type, file_base_container_type >
    index_container_type;

    //-------------------------------------------------------------------
    /// \brief Index of all SDGT.x files for a directory
    //-------------------------------------------------------------------
    index_container_type        p_index;

    void find_ifo_type( const Query& SearchParams,
                        QueryResults& Results,
                        const file_base_container_type& IfoTypeInfo,
                        const std::string& Extension ) const;

    void find_time_intervals( const Query& SearchParams,
                              QueryResults& Results,
                              file_base_container_type::const_iterator IfoTypeInfo,
                              const std::string& Extension ) const;

    index_container_type& index( );

    const index_container_type& index( ) const;
  };

  inline lHash::index_container_type& lHash::
  index( )
  {
    return p_index;
  }

  inline const lHash::index_container_type& lHash::
  index( ) const
  {
    return p_index;
  }

  struct lIsFilePrivateData
    : public diskCache::FilenameHandler::IsFileData::PrivateData
  {
    struct new_entry
    {
      std::string       s_ifotype;
      std::string       s_extension;
      INT_4U            s_gps_start;
      INT_4U            s_dt;
    };

    lIsFilePrivateData( )
      : s_new_entries( lHash::dir_info_type( ) ),
        s_reset( false )
    {
    }

    ~lIsFilePrivateData( )
    {
    }

    inline
    void NewEntry( const std::string& IFOType,
                   INT_4U GPSStart,
                   INT_4U DT,
                   const std::string& Extension )
    {
      if ( s_new_entries.Overlapping( IFOType,
                                      GPSStart,
                                      DT,
                                      Extension )
           == false )
      {
        s_new_entries.Add( IFOType,
                           GPSStart,
                           DT,
                           Extension );
      }
    }

    //-------------------------------------------------------------------
    /// \brief Add new entries to hash once directory has been read.
    ///
    /// After reading the contents of a directory:
    /// <ul>
    /// <li> Verify any new files as not being in conflict.
    ///     <ul>
    ///     <li> Generate appropriate error message for conflicting data.
    ///     <li> Add valid data to searchable hash
    ///     </ul>
    /// <li> Calculate actual number of files that have been added.
    /// <li> Calcuate actual number of files that have been removed.
    /// </ul>
    ///
    /// \param[in] Dir
    ///     Object containing the current directory information.
    //-------------------------------------------------------------------
    virtual void
    OnDirectoryClose( General::SharedPtr< DirectoryInfo > Dir )
    {
      s_added_file_count = s_current_lhash->Add( s_new_entries,
                                                 s_conflict_messages );
      //----------------------------------------------------------------
      // Calculate the number of files removed
      // removed  = previous_count - common_files
      //----------------------------------------------------------------
      const INT_4U    pc = ( ( s_previous_lhash )
                             ? s_previous_lhash->FileCount( )
                             : 0 );
      if ( pc )
      {
        s_removed_file_count =
          pc                                      // previous_count
          - ( s_current_lhash->FileCount( )
              - s_added_file_count )              // common_files
          ;
      }
      else
      {
        s_removed_file_count = 0;
      }
      //----------------------------------------------------------------
      // 
      //----------------------------------------------------------------
      if ( s_reset )
      {
        Dir->TimeModifiedReset( );
      }
    }

    typedef General::SharedPtr< diskCache::FilenameHandler::Hash >
    hash_type;

    typedef lHash                       new_entry_container_type;

    hash_type                   s_previous_hash;
    hash_type                   s_current_hash;
    lHash*                      s_previous_lhash;
    lHash*                      s_current_lhash;
    new_entry_container_type    s_new_entries;
    bool                        s_reset;
  };

#if ENABLE_MOUNT_POINT_CACHE
  //=====================================================================
  //
  //=====================================================================
  class lMountPointCache
    : public MountPointCache
  {
  public:
    typedef MountPointCache::Info::directory_type directory_type;

    class lInfo
      : public MountPointCache::Info
    {
    public:
      typedef Info::directory_type directory_type;

      //-----------------------------------------------------------------
      /// \brief constructor
      //-----------------------------------------------------------------
      lInfo( directory_type DirInfo,
             lHash* FileInfo );

      //-----------------------------------------------------------------
      /// \brief Destructor
      //-----------------------------------------------------------------
      virtual ~lInfo( );

      bool CheckForConflicts( ) const;

      void CheckForConflicts( bool Value );

      const lHash& Hash( ) const;

      lHash& Hash( );

    private:
      bool              m_check_for_conflicts;
      lHash*            m_file_info;
    };

    //-------------------------------------------------------------------
    /// Hash of known subdirectories with "IFO-Type" data for a particular
    /// MOUNT_PT entry:
    /// "IFO-Type-" <---> vector< DirectoryInfo* > 
    ///
    /// \warning
    ///     Don't ever delete any of these DirectoryInfo* objects: 
    ///     these are shallow pointers!!!
    //-------------------------------------------------------------------
    typedef General::unordered_map< std::string,
                                      std::vector< directory_type > >
    IFOTypeHash;

    //-------------------------------------------------------------------
    //-------------------------------------------------------------------
    typedef General::unordered_map< std::string, IFOTypeHash >
    ExtHash;

    //-------------------------------------------------------------------
    /// \brief Destructor
    //-------------------------------------------------------------------
    ~lMountPointCache( );

    //-------------------------------------------------------------------
    /// \brief Search for information within the cache
    ///	
    /// \param[in] SearchParams
    ///     The search parameters describing the data to find.
    /// \param[out] Results
    ///     The data found.
    //-------------------------------------------------------------------
    virtual void Find( const Query& SearchParams,
		       QueryResults& Results );

    //-------------------------------------------------------------------
    /// \brief Register cache information
    //-------------------------------------------------------------------
    virtual void Register( General::SharedPtr< diskCache::FilenameHandler::Hash >
                           DirectoryInfoHash );

    //-------------------------------------------------------------------
    /// \brief Unregister cache information
    //-------------------------------------------------------------------
    virtual void Unregister( General::SharedPtr< diskCache::FilenameHandler::Hash >
                             DirectoryInfoHash );

    //-------------------------------------------------------------------
    /// \brief Check the state of emptiness.
    //-------------------------------------------------------------------
    virtual bool empty( ) const;

  private:
    //-------------------------------------------------------------------
    /// \brief Protect an instance of the object in multi-threaded env.
    ///
    /// This variable allows for synconization to occur when this
    /// an instance of this object is being referenced in a
    /// multi-threaded environment.
    //-------------------------------------------------------------------
    General::ReadWriteLock::Sync         m_baton;

    ExtHash             m_cache;

    //-------------------------------------------------------------------
    /// \brief Search for information within the cache
    ///	
    /// \param[in] SearchParams
    ///     The search parameters describing the data to find.
    /// \param[out] Results
    ///     The data found.
    //-------------------------------------------------------------------
    void find_ifo_type( const Query& SearchParams,
                        QueryResults& Results,
                        const IFOTypeHash& IfoTypeInfo );

    //-------------------------------------------------------------------
    /// \brief Register directory information.
    //-------------------------------------------------------------------
    void register_dir_info( const std::string& Extension,
                            const std::string& IFO_Type,
                            directory_type Directory );

    //-------------------------------------------------------------------
    /// \brief Unregister directory information.
    //-------------------------------------------------------------------
    void unregister_dir_info( const std::string& Extension,
                              const std::string& IFO_Type,
                              directory_type Directory );

  };

  //=====================================================================
  //=====================================================================
  inline bool lMountPointCache::lInfo::
  CheckForConflicts( ) const
  {
    return m_check_for_conflicts;
  }
  
  inline void lMountPointCache::lInfo::
  CheckForConflicts( bool Value )
  {
    m_check_for_conflicts = Value;
  }
  
  const lHash& lMountPointCache::lInfo::
  Hash( ) const
  {
    if ( m_file_info )
    {
      return *m_file_info;
    }
    throw std::bad_cast( );
  }

  lHash& lMountPointCache::lInfo::
  Hash( )
  {
    if ( m_file_info )
    {
      return *m_file_info;
    }
    throw std::bad_cast( );
  }

  //=====================================================================
  //=====================================================================
  inline void lMountPointCache::
  register_dir_info( const std::string& Extension,
                     const std::string& IFO_Type,
                     directory_type Directory )
  {
    // operator[] will insert hash entry if it doesn't exist
    ExtHash::mapped_type::mapped_type&
      known_dirs( m_cache[ Extension ][ IFO_Type ] );
   
    if ( std::find( known_dirs.begin( ),
                    known_dirs.end( ),
                    Directory )
         == known_dirs.end( ) )
    {
      known_dirs.push_back( Directory );
    }
  }

  //=====================================================================
  //=====================================================================
  inline void lMountPointCache::
  unregister_dir_info( const std::string& Extension,
                       const std::string& IFO_Type,
                       directory_type Directory )
  {
    ExtHash::iterator pos_ext = m_cache.find( Extension );

    if ( pos_ext != m_cache.end( ) )
    {
      ExtHash::mapped_type::iterator
        pos_ifo_type = pos_ext->second.find( IFO_Type );

      if ( pos_ifo_type != pos_ext->second.end( ) )
      {
        //----------------------------------------------------------------
        // Remove the directory from the list if it is found
        //----------------------------------------------------------------
        ExtHash::mapped_type::mapped_type::iterator
          pos_directory = std::find( pos_ifo_type->second.begin( ),
                                     pos_ifo_type->second.end( ),
                                     Directory );
        if ( pos_directory != pos_ifo_type->second.end( ) )
        {
          pos_ifo_type->second.erase( pos_directory );
        }
        if ( pos_ifo_type->second.empty( ) )
        {
          //-------------------------------------------------------------
          // Removed the last ifo_type entry, so cleanup
          //-------------------------------------------------------------
          pos_ext->second.erase( pos_ifo_type );
        }
      }
      if ( pos_ext->second.empty( ) )
      {
        //---------------------------------------------------------------
        // Removed the last extension entry, so cleanup
        //---------------------------------------------------------------
        m_cache.erase( pos_ext );
      }
    }
  }

  inline void lMountPointCache::
  find_ifo_type( const Query& SearchParams,
                 QueryResults& Results,
                 const IFOTypeHash& IfoTypeInfo )
  {
    //-------------------------------------------------------------------
    // Get a copy of the search parameters
    //-------------------------------------------------------------------
    Query sp( SearchParams );

    if ( SearchParams.Criteria( ) & Query::CRITERIA_BIT_IFO_TYPE_LIST )
    {
      for ( Query::ifo_type_list_type::const_iterator
              cur = SearchParams.IFOTypeList( ).begin( ),
              last = SearchParams.IFOTypeList( ).end( );
            cur != last;
            ++cur )
      {
        size_t pos = cur->find_first_of( "-" );
        sp.IFO( cur->substr( 0, pos + 1 ) );
        sp.Type( cur->substr( pos + 1 ) );

        find_ifo_type( sp, Results, IfoTypeInfo );
      }
      return;
    }
    else if ( SearchParams.Criteria( ) & Query::CRITERIA_BIT_IFO_TYPE )
    {
      const std::string& ifo_type( SearchParams.IFOType( ) );
      size_t pos = ifo_type.find_first_of( "-" );
      sp.IFO( ifo_type.substr( 0, pos + 1 ) );
      sp.Type( ifo_type.substr( pos + 1 ) );
      
      find_ifo_type( sp, Results, IfoTypeInfo );
      return;
    }
    else if ( ( SearchParams.Criteria( )
                & ( Query::CRITERIA_BIT_IFO
                    | Query::CRITERIA_BIT_TYPE ) ) ==
              ( Query::CRITERIA_BIT_IFO
                | Query::CRITERIA_BIT_TYPE ) )
    {
      const IfoTypeMatch match( SearchParams.IFO( ),
                                SearchParams.Type( ) );
    
      for( IFOTypeHash::const_iterator
             cur = IfoTypeInfo.begin( ),
             last = IfoTypeInfo.end();
           cur != last;
           ++cur )
      {
        // Examine "IFO-Type-" key of the map: 
        // if it matches searched criteria
        // then get all known data from all registered directories 
        // for this "IFO-Type-"
        if ( match( cur->first ) )
        {
          const IFOTypeHash::mapped_type& directories( cur->second );
   
          for ( IFOTypeHash::mapped_type::const_iterator
                  cur_dir = directories.begin( ),
                  last_dir = directories.end( );
                cur_dir != last_dir;
                ++cur_dir )
          {
            sp.IFOType( cur->first );

            ( *cur_dir )->Find( sp, Results );
          }
        } // matched searched ifo and type
      } // step through all registered "IFO-Type-" keys of the mount point entry
    }
  }
#endif /* ENABLE_MOUNT_POINT_CACHE */

  //--------------------------------------------------------------------
  // 
  //--------------------------------------------------------------------
  class qrConflicts
    : public QueryResultsTimeRange
  {
  public:
    typedef diskCache::FilenameHandler::dir_info_type dir_info_type;
    typedef TimeHash::KnownIntervals source_type;
    typedef ::conflict_type conflict_type;

    class conflict_range
    {
    public:
      inline conflict_range( INT_4U Start,
                             INT_4U End,
                             INT_4U DT )
        : m_start( Start ),
          m_end( End ),
          m_dt( DT )
      {
      }

      inline void write( std::ostream& Stream ) const
      {
        Stream << "([" << m_start
               << "," << m_end
               << "):" << m_dt
               << ")"
          ;
      }

      inline bool empty( ) const
      {
        return ( m_start == m_end );
      }

      inline bool operator==( const conflict_range& CR ) const
      {
        return ( ( m_start == CR.m_start )
                 && ( m_end == CR.m_end )
                 && ( m_dt == CR.m_dt ) );
      }

      inline void intersection( const conflict_range& RHS )
      {
        if ( ( RHS.m_start >= m_end )
             || ( RHS.m_end <= m_start ) )
        {
          m_start = 0;
          m_end = 0;
          return;
        }
        //-----------------------------------------------------------------
        // Shrink the interval size to only be what is in common
        //-----------------------------------------------------------------
        if ( RHS.m_start > m_start )
        {
          m_start = ifloor( RHS.m_start, m_start, m_dt );
        }
        if ( RHS.m_end < m_end )
        {
          m_end = iceil( RHS.m_end, m_end, m_dt );
        }
      }

      INT_4U	m_start;
      INT_4U	m_end;
      INT_4U	m_dt;
    };

    qrConflicts( dir_info_type SourceDir,
                 source_type& Source );


    void Finalize( const Query& SearchParams,
                   diskCache::FilenameHandler::IsFileData::conflict_container_type& Conflicts );
    
  private:
    conflict_type       m_conflicts;
    source_type&        m_source;
    dir_info_type       m_source_dir;
  };

  qrConflicts::
  qrConflicts( dir_info_type SourceDir,
               source_type& Source )
    : QueryResultsTimeRange(  QueryResults::FORMAT_CONFLICT, true ),
      m_source( Source ),
      m_source_dir( SourceDir )
  {
  }

  //--------------------------------------------------------------------
  // 
  //--------------------------------------------------------------------
  static diskCache::SDGTxFilenameHandler	handler;

} // namespace - anonymous

namespace std
{
  std::ostream&
  operator<<( std::ostream& Stream, const qrConflicts::conflict_range& Data )
  {
    Data.write( Stream );
    return Stream;
  }

  std::ostream&
  operator<<( std::ostream& Stream, const lHash& Data )
  {
    return Data.Write( Stream );
  }

}

namespace diskCache
{
  FilenameHandler::Hash* SDGTxFilenameHandler::
  CreateHash( dir_info_type Value ) const
  {
    Hash* retval = new lHash( Value );
    return retval;
  }

#if ENABLE_MOUNT_POINT_CACHE
  MountPointCache* SDGTxFilenameHandler::
  CreateMountPointCache( ) const
  {
    return new lMountPointCache;
  }
#endif /* ENABLE_MOUNT_POINT_CACHE */

  bool SDGTxFilenameHandler::
  IsFile( const std::string& Filename, FilenameHandler::IsFileData& Data ) const
  {
    RegexMatch	match( SDGTx_MATCH_SIZE );

    try
    {
      bool m = match.match( SDGTx_RegEx, Filename.c_str( ) );

      if ( m == false )
      {
	return false;
      }
    }
    catch( ... )
    {
      return false;
    }

    std::string         ext( match.getSubString( SDGTx_x_OFFSET ) );
    std::istringstream	g_str( match.getSubString( SDGTx_G_OFFSET ) );
    std::istringstream	t_str( match.getSubString( SDGTx_T_OFFSET ) );
    INT_4U	g = 0;
    INT_4U	t = 0;
			       
    g_str >> g;
    t_str >> t;

    std::string	ifo_type( match.getSubString( SDGTx_S_OFFSET ) );
    ifo_type += ::SEPERATOR( );
    ifo_type += match.getSubString( SDGTx_D_OFFSET );
    ifo_type += ::SEPERATOR( );

    try
    {
      //---------------------------------------------------------
      // This filename is looking like it meets all the
      // criteria for a new file.
      //---------------------------------------------------------

      //---------------------------------------------------------
      /// Obtain the private state data
      //---------------------------------------------------------
      IsFileData::private_data_container_type::iterator
        pd = Data.s_private_data.find( SDGTxFilenameHandler::Key( ) );
      lIsFilePrivateData*       lpd;

      if ( pd == Data.s_private_data.end( ) )
      {
        Data.s_private_data[ SDGTxFilenameHandler::Key( ) ]
          = IsFileData::private_data_container_type::
          mapped_type( new lIsFilePrivateData );
        pd = Data.s_private_data.find( SDGTxFilenameHandler::Key( ) );
        lpd = dynamic_cast< lIsFilePrivateData* >( pd->second.get( ) );
        //-------------------------------------------------------
        // Locate the hash info for the previous contents of
        //     the directory.
        //-------------------------------------------------------
        {
          const DirectoryInfo::index_variable_ro_type p = Data.s_previous_view.Index( );

          for ( DirectoryInfo::index_type::const_iterator
                  cur = p.Var( ).begin( ),
                  last = p.Var( ).end( );
                cur != last;
                ++cur )
          {
            if ( cur->first == SDGTxFilenameHandler::Key( ) )
            {
              lpd->s_previous_hash = cur->second;
              break;
            }
          }
          
        }
        //-------------------------------------------------------
        // Locate the hash info for the current contents of
        //     the directory.
        //-------------------------------------------------------
        {
          DirectoryInfo::index_variable_rw_type p = Data.s_current_view->Index( );

          for ( DirectoryInfo::index_type::iterator
                  cur = p.Var( ).begin( ),
                  last = p.Var( ).end( );
                cur != last;
                ++cur )
          {
            if ( cur->first == SDGTxFilenameHandler::Key( ) )
            {
              //-------------------------------------------------
              // Entry exists
              //-------------------------------------------------
              if ( cur->second.get( ) )
              {
                //-----------------------------------------------
                // Allocate resources to store the hash info
                //-----------------------------------------------
                cur->second.reset( SDGTxFilenameHandler::Key( )->CreateHash( Data.s_current_view ) );
              }
              lpd->s_current_hash = cur->second;
              break;
            }
          }
          if ( lpd->s_current_hash.get( ) == 0 )
          {
            //---------------------------------------------------
            // Entry does not exist. Create it all.
            //---------------------------------------------------
            lpd->s_current_hash.reset( SDGTxFilenameHandler::Key( )->CreateHash( Data.s_current_view ) );
            DirectoryInfo::index_type::value_type
              v( SDGTxFilenameHandler::Key( ),
                 lpd->s_current_hash );

            p.Var( ).push_back( v );
          }
        }
        lpd->s_previous_lhash = dynamic_cast< lHash* >( lpd->s_previous_hash.get( ) );
        lpd->s_current_lhash = dynamic_cast< lHash* >( lpd->s_current_hash.get( ) );
      }
      else
      {
        lpd = dynamic_cast< lIsFilePrivateData* >( pd->second.get( ) );
      }
      //---------------------------------------------------------
      // Check if new
      //---------------------------------------------------------
      if ( ( lpd->s_previous_lhash == 0 )
           || ( lpd->s_previous_lhash->Find( ifo_type, g, t, ext ) == false ) )
      {
        //-------------------------------------------------------
        // Any new entry needs to be validated only after all
        // of the entries have been read from the directory.
        //-------------------------------------------------------
        lpd->NewEntry( ifo_type, g, t, ext );
        return true;
      }
      //---------------------------------------------------------
      // Add entry to list of managed files since it is not
      // in conflict with existing data.
      //---------------------------------------------------------
      lpd->s_current_lhash->Add( ifo_type, g, t, ext );
      return true;
    }
    catch( ... )
    {
      return false;
    }
  }

  bool SDGTxFilenameHandler::
  Register( )
  {
    return FilenameHandler::RegisterHandler( ::handler );
  }

  bool SDGTxFilenameHandler::
  Unregister( )
  {
    return FilenameHandler::UnregisterHandler( ::handler );
  }

  FilenameHandler* SDGTxFilenameHandler::
  Key( )
  {
    return &::handler;
  }

} /* namespace diskCache */

namespace
{
#if ENABLE_MOUNT_POINT_CACHE
  //=====================================================================
  //=====================================================================
  lMountPointCache::lInfo::
  lInfo( directory_type DirInfo,
         lHash* FileInfo )
    : Info( diskCache::SDGTxFilenameHandler::Key( ),
            DirInfo ),
      m_check_for_conflicts( true ),
      m_file_info( FileInfo )
  {
  }

  //---------------------------------------------------------------------
  // Release all resouces back to the system
  //---------------------------------------------------------------------
  lMountPointCache::lInfo::
  ~lInfo( )
  {
  }

  //---------------------------------------------------------------------
  /// Release all resouces back to the system
  //---------------------------------------------------------------------
  lMountPointCache::
  ~lMountPointCache( )
  {
  }

  //---------------------------------------------------------------------
  //---------------------------------------------------------------------
  void lMountPointCache::
  Find( const Query& SearchParams,
        QueryResults& Results )
  {
    General::ReadWriteLock l( m_baton,
                              General::ReadWriteLock::READ,
                              __FILE__, __LINE__ );

    static const Query::bit_flag_type   mask_1 =
      ( Query::CRITERIA_BIT_EXT
        | Query::CRITERIA_BIT_IFO
        | Query::CRITERIA_BIT_TYPE );

    static const Query::bit_flag_type   mask_2 =
      ( Query::CRITERIA_BIT_EXT
        | Query::CRITERIA_BIT_IFO_TYPE );

    static const Query::bit_flag_type   mask_3 =
      ( Query::CRITERIA_BIT_EXT
        | Query::CRITERIA_BIT_IFO_TYPE_LIST );

    const std::string&
      ext( ( SearchParams.Criteria( ) & Query::CRITERIA_BIT_EXT )
           ? SearchParams.Extension( )
           : GLOB_ALL );

    const Query::bit_flag_type
      criteria = SearchParams.Criteria( ) | Query::CRITERIA_BIT_EXT;

    if ( ( ( criteria & mask_1 ) != mask_1 )
         && ( ( criteria & mask_2 ) != mask_2 )
         && ( ( criteria & mask_3 ) != mask_3 ) )
    {
      //-----------------------------------------------------------------
      // Can only answer queries involving the minimum of file extension,
      // ifo and type
      //-----------------------------------------------------------------
      return;
    }

    if ( ext == GLOB_ALL )
    {
      //-----------------------------------------------------------------
      // Get a copy of the search parameters
      //-----------------------------------------------------------------
      Query sp( SearchParams );

      //-----------------------------------------------------------------
      // Loop over all extensions
      //-----------------------------------------------------------------
      for ( ExtHash::const_iterator
              cur = m_cache.begin( ),
              last = m_cache.end( );
            cur != last;
            ++cur )
      {
        //---------------------------------------------------------------
        // Reset the query to only look within this extension
        //---------------------------------------------------------------
        sp.Extension( cur->first );

        find_ifo_type( sp, Results, cur->second );
      }
    }
    else
    {
      //-----------------------------------------------------------------
      // Find the specific extension
      //-----------------------------------------------------------------
      ExtHash::const_iterator pos = m_cache.find( ext );

      if ( pos != m_cache.end( ) )
      {
        find_ifo_type( SearchParams, Results, pos->second );
      }
    }
  }

  //---------------------------------------------------------------------
  //---------------------------------------------------------------------
  void lMountPointCache::
  Register( General::SharedPtr< diskCache::FilenameHandler::Hash >
            DirectoryInfoHash )
  {
    lHash*      hash( dynamic_cast< lHash* >( DirectoryInfoHash.get( ) ) );

    if ( hash )
    {
      General::ReadWriteLock l( m_baton,
                                General::ReadWriteLock::WRITE,
                                __FILE__, __LINE__ );
      lHash::dir_info_type      dir_info;

      //-----------------------------------------------------------------
      // Register all of this data
      // Loop over the file extensions
      //-----------------------------------------------------------------
      for( lHash::index_container_type::const_iterator
             cur_ext = hash->index( ).begin( ),
             last_ext = hash->index( ).end( );
           cur_ext != last_ext;
           ++cur_ext )
      {
        //---------------------------------------------------------------
        // Loop over the ifo_type
        //---------------------------------------------------------------
        for ( lHash::file_base_container_type::const_iterator
                cur_site_desc = cur_ext->second.begin( ),
                last_site_desc = cur_ext->second.end( );
              cur_site_desc != last_site_desc;
              ++cur_site_desc )
        {
          if ( dir_info = hash->m_directory_info_ref.lock( ) )
          {
            register_dir_info( cur_ext->first,
                               cur_site_desc->first,
                               dir_info );
          }
        }
      }
    }
  }

  void lMountPointCache::
  Unregister( General::SharedPtr< diskCache::FilenameHandler::Hash >
              DirectoryInfoHash )
  {
    lHash*      hash( dynamic_cast< lHash* >( DirectoryInfoHash.get( ) ) );

    if ( hash )
    {
      lHash::dir_info_type      dir_info;

      General::ReadWriteLock l( m_baton,
                                General::ReadWriteLock::WRITE,
                                __FILE__, __LINE__ );

      //-----------------------------------------------------------------
      // Register all of this data
      // Loop over the file extensions
      //-----------------------------------------------------------------
      for( lHash::index_container_type::const_iterator
             cur_ext = hash->index( ).begin( ),
             last_ext = hash->index( ).end( );
           cur_ext != last_ext;
           ++cur_ext )
      {
        //---------------------------------------------------------------
        // Loop over the ifo_type
        //---------------------------------------------------------------
        for ( lHash::file_base_container_type::const_iterator
                cur_site_desc = cur_ext->second.begin( ),
                last_site_desc = cur_ext->second.end( );
              cur_site_desc != last_site_desc;
              ++cur_site_desc )
        {
          if ( dir_info = hash->m_directory_info_ref.lock( ) )
          {
            unregister_dir_info( cur_ext->first,
                                 cur_site_desc->first,
                                 dir_info );
          }
        }
      }
    }
  }

  //---------------------------------------------------------------------
  //---------------------------------------------------------------------
  bool lMountPointCache::
  empty( ) const
  {
    return m_cache.empty( );
  }
#endif /* ENABLE_MOUNT_POINT_CACHE */

  //=====================================================================
  // lHash
  //=====================================================================
  lHash::
  lHash( dir_info_type Value )
    : Hash( Value )
  {
  }

  lHash::
  ~lHash( )
  {
  }

  //-------------------------------------------------------------
  // This method add information into the hash data.
  // It will create any nodes that are needed.
  // It is possible for an insertion to result in
  // the collapsing of data nodes as holes are filled.
  //-------------------------------------------------------------
  void lHash::
  Add( const std::string& IFOType,
       INT_4U GPSStart,
       INT_4U DT,
       const std::string& Extension )
  {
    time_data_type& td( index( )[ Extension][ IFOType ] );
    if( td.get( ) == NULL )
    {
      td.reset( new time_data_type::element_type );
    }
    td->Add( GPSStart, DT );
  }

  INT_4U lHash::
  Add( const lHash& Source,
       conflict_container_type& Conflicts )
  {
    INT_4U retval( 0 );

    Query                         q;

    q.FNHandler( diskCache::SDGTxFilenameHandler::Key( ) );
    dir_info_type       d( m_directory_info_ref.lock( ) );
    if ( ! d )
    {
      return retval;
    }

    q.MountPoint( d->GetMountPointPath( ).Var( ) );
    //-----------------------------------------------------------
    // Loop over the file extensions
    //-----------------------------------------------------------
    for ( index_container_type::const_iterator
            ext_cur = Source.index( ).begin( ),
            ext_last = Source.index( ).end( );
          ext_cur != ext_last;
          ++ext_cur )
    {
      q.Extension( ext_cur->first );
      for ( file_base_container_type::const_iterator
              ifo_type_cur = ext_cur->second.begin( ),
              ifo_type_last = ext_cur->second.end( );
            ifo_type_cur != ifo_type_last;
            ++ifo_type_cur )
      {
        //-------------------------------------------------------
        // Query for conflicts
        //-------------------------------------------------------
        q.IFOType( ifo_type_cur->first );

        INT_4U                          start;
        INT_4U                          stop;
        Query                           kiq;
        TimeHash::KnownIntervals        ki;

        ifo_type_cur->second->Find( kiq, ki );
        start = ki.Ranges( ).begin( )->first;
        stop = ki.Ranges( ).begin( )->second.mStopTime;
        for ( TimeHash::KnownIntervals::ranges_type::const_iterator
                ki_cur = ki.Ranges( ).begin( ),
                ki_last = ki.Ranges( ).end( );
              ki_cur != ki_last;
              ++ki_cur )
        {
          if ( ki_cur->second.mStopTime > stop )
          {
            stop = ki_cur->second.mStopTime;
          }
        }
        q.TimeRange( start, stop );
        {
          qrConflicts   results( d, ki );

          MountPointHash::Find( q, results );

          results.Finalize( q, Conflicts );

          time_data_type& e( index( )[ ext_cur->first ][ ifo_type_cur->first ] );
          if ( e.get( ) == NULL )
          {
            e.reset( new time_data_type::element_type );
          }
          e->Add( ki );
          retval += ki.CountFiles( );
        }
      }

    }
    return retval;
  }

  //-------------------------------------------------------------
  // This function with search the current hash information
  // based on the information supplied.
  // If no match is found, then false is returned.
  //-------------------------------------------------------------
  bool lHash::
  Find( const std::string& IFOType,
        INT_4U GPSStart,
        INT_4U DT,
        const std::string& Extension ) const
  {
    //-----------------------------------------------------------
    // Search for the Extension
    //-----------------------------------------------------------
    lHash::index_container_type::const_iterator
      ext_pos = index( ).find( Extension );
    if ( ext_pos == index( ).end( ) )
    {
      //---------------------------------------------------------
      // Does not exist in cached data
      //---------------------------------------------------------
      return false;
    }
    //-----------------------------------------------------------
    // Search for the IFOType
    //-----------------------------------------------------------
    file_base_container_type::const_iterator
      ifotype_pos = ext_pos->second.find( IFOType );
    if ( ( ifotype_pos == ext_pos->second.end( ) )
         || ( ! ifotype_pos->second ) )
    {
      //---------------------------------------------------------
      // Does not exist in cached data
      //---------------------------------------------------------
      return false;
    }
    //-----------------------------------------------------------
    // Search for the GPS start time with the delta time
    //-----------------------------------------------------------
    return ifotype_pos->second->Find( GPSStart, DT );
  }

  void lHash::
  Find( const Query& SearchParams,
        QueryResults& Results ) const
  {
    const Query::bit_flag_type  criteria( SearchParams.Criteria( ) );

    if ( criteria & Query::CRITERIA_BIT_EXT )
    {
      index_container_type::const_iterator
        pos = index( ).find( SearchParams.Extension( ) );
      if ( pos != index( ).end( ) )
      {
        find_ifo_type( SearchParams, Results, pos->second, pos->first );
      }
    }
    else
    {
      Query     sp( SearchParams );

      for ( index_container_type::const_iterator
              cur = index( ).begin( ),
              last = index( ).end( );
            cur != last;
            ++cur )
      {
        sp.Extension( cur->first );

        find_ifo_type( SearchParams, Results, cur->second, cur->first );
      }
    }
  }

  //-------------------------------------------------------------
  //-------------------------------------------------------------
  bool lHash::
  Overlapping( const std::string& IFOType,
               INT_4U GPSStart,
               INT_4U GPSEnd,
               const std::string& Extension ) const
  {
    //-----------------------------------------------------------
    // Search for the Extension
    //-----------------------------------------------------------
    lHash::index_container_type::const_iterator
      ext_pos = index( ).find( Extension );
    if ( ext_pos == index( ).end( ) )
    {
      //---------------------------------------------------------
      // Does not exist in cached data
      //---------------------------------------------------------
      return false;
    }
    //-----------------------------------------------------------
    // Search for the IFOType
    //-----------------------------------------------------------
    file_base_container_type::const_iterator
      ifotype_pos = ext_pos->second.find( IFOType );
    if ( ( ifotype_pos == ext_pos->second.end( ) )
         || ( ! ifotype_pos->second ) )
    {
      //---------------------------------------------------------
      // Does not exist in cached data
      //---------------------------------------------------------
      return false;
    }
    //-----------------------------------------------------------
    // Search for the GPS start time with the delta time
    //-----------------------------------------------------------
    return ifotype_pos->second->Overlapping( GPSStart, GPSEnd );
  }

  size_t lHash::
  FileCount( ) const
  {
    size_t      retval( 0 );

    for( index_container_type::const_iterator
           cur_ext = index( ).begin( ),
           last_ext = index( ).end( );
         cur_ext != last_ext;
         ++cur_ext )
    {
      for ( file_base_container_type::const_iterator
              cur = cur_ext->second.begin( ),
              last = cur_ext->second.end( );
            cur != last;
            ++cur )
      {
        if ( cur->second )
        {
          retval += cur->second->Count( );
        }
      }
    }
    return retval;
  }

  void lHash::
  Read( CacheFile& Stream )
  {
    switch ( Stream.Version( ) )
    {
    case 0x0101:
      {
        CacheFile::size_type    ext_size;

        Stream >> ext_size;

        index( ).erase( index( ).begin( ), index( ).end( ) );

        while( ext_size-- )
        {
          std::string           ext_name;
          CacheFile::size_type  size;

          Stream >> ext_name
                 >> size
            ;

          file_base_container_type& pos
            = index( )[ ext_name ];

          while ( size-- != 0 )
          {
            file_base_container_type::key_type      key;
            std::auto_ptr< file_base_container_type::mapped_type::element_type >
              value;

            Stream >> key
                   >> value
              ;
            pos[ key ].reset( value.release( ) );
          }
      
        }
        //---------------------------------------------------------------
      }
      break;
    case 0x0100:
    case 0x0000:
      {
        //---------------------------------------------------------------
        // Get the .gwf entry; create it if it does not exist
        //---------------------------------------------------------------
        CacheFile::size_type    size;

        Stream >> size;

        if ( size > 0 )
        {
          file_base_container_type&
            gwf_pos = index( )[ ".gwf" ];

          while ( size-- != 0 )
          {
            file_base_container_type::key_type      key;
            std::auto_ptr< file_base_container_type::mapped_type::element_type >
              value;
            
            Stream >> key
                   >> value
              ;
            gwf_pos[ key ].reset( value.release( ) );
          }
        }
      
      }
      break;
    default:
      {
        std::ostringstream        msg;

        msg
          << "anonymous::SDGTxFilenameHandler::lHash::Read Unsupported version: "
          << Stream.Version( )
          ;
        throw std::runtime_error(  msg.str( ) );
      }
      break;
    } // switch - version
  }

  std::ostream& lHash::
  Write( std::ostream& Stream ) const
  {
    //-----------------------------------------------------------
    // Loop over extensions
    //-----------------------------------------------------------
    dir_info_type       di( m_directory_info_ref.lock( ) );
    std::string         path( di->GetPath( ).Var( ) );

    for ( index_container_type::const_iterator
            ext_cur = index( ).begin( ),
            ext_last = index( ).end( );
          ext_cur != ext_last;
          ++ext_cur )
    {
      for ( file_base_container_type::const_iterator
              ifotype_cur = ext_cur->second.begin( ),
              ifotype_last = ext_cur->second.end( );
            ifotype_cur != ifotype_last;
            ++ifotype_cur )
      {
        if ( ( ! ifotype_cur->second )
             || ( ifotype_cur->second->empty( ) ) )
        {
          continue;
        }
        Query                           sp;     // Search parameters
        TimeHash::KnownIntervals        ki;     // Known intervals

        ifotype_cur->second->Find( sp, ki );

        for ( TimeHash::KnownIntervals::ranges_type::const_iterator
                range_cur = ki.Ranges( ).begin( ),
                range_last = ki.Ranges( ).end( );
              range_cur != range_last;
              ++range_cur )
        {
          INT_4U        start = range_cur->first;
          INT_4U        stop = range_cur->second.mStopTime;
          INT_4U        dt = range_cur->second.mFileDt;

          while( start < stop )
          {
            Stream << std::endl << "\t" << path << "/"<< ifotype_cur->first
                   << start << "-" << dt
                   << ext_cur->first
              ;
            start += dt;
          }
        }
      }
#if 0
    //-----------------------------------------------------------
    // Search for the GPS start time with the delta time
    //-----------------------------------------------------------
    return ifotype_pos->second->Find( GPSStart, DT );
#endif /* 0 */
    
    }
    return Stream;
  }

  void lHash::
  Write( CacheFile& Stream ) const
  {
    switch( Stream.Version( ) )
    {
    case 0x0101:
      {
        CacheFile::size_type    ext_size = index( ).size( );

        Stream << ext_size;

        for( index_container_type::const_iterator
               cur_ext = index( ).begin( ),
               last_ext = index( ).end( );
             cur_ext != last_ext;
             ++cur_ext )
        {
          CacheFile::size_type    size = cur_ext->second.size( );

          Stream
            << cur_ext->first   // Output the extension
            << size             // Output the number of entries
            ;

          for ( file_base_container_type::const_iterator
                  cur = cur_ext->second.begin( ),
                  last = cur_ext->second.end( );
                cur != last;
                ++cur )
          {
            Stream << cur->first        // Output the site/desc info
                   << cur->second
              ;
            
          }
        }
      }
      break;
    case 0x0100:
    case 0x0000:
      {
        //---------------------------------------------------------------
        // Locate the .gwf entry
        //---------------------------------------------------------------
        index_container_type::const_iterator
          gwf_pos = index( ).find( ".gwf" );

        CacheFile::size_type    size = ( ( gwf_pos == index( ).end( ) )
                                         ? 0
                                         : gwf_pos->second.size( ) );

        Stream << size;

        if ( size > 0 )
        {
          for ( file_base_container_type::const_iterator
                  cur = gwf_pos->second.begin( ),
                  last = gwf_pos->second.end( );
                cur != last;
                ++cur )
          {
            Stream << cur->first
                   << cur->second
              ;
            
          }
        }
      }
      break;
    default:
      {
        std::ostringstream        msg;

        msg
          << "anonymous::SDGTxFilenameHandler::lHash::Write Unsupported version: "
          << Stream.Version( )
          ;
        throw std::runtime_error(  msg.str( ) );
      }
      break;
    }
  }

  void lHash::
  find_ifo_type( const Query& SearchParams,
                 QueryResults& Results,
                 const file_base_container_type& IfoTypeInfo,
                 const std::string& Extension ) const
  {
    const Query::bit_flag_type criteria( SearchParams.Criteria( ) );

    if ( criteria & ( Query::CRITERIA_BIT_IFO_TYPE
                      | Query::CRITERIA_BIT_IFO_TYPE_ABSOLUTE ) )
    {
      //-----------------------------------------------------------------
      // Find exact match
      //-----------------------------------------------------------------
      file_base_container_type::const_iterator
        pos_ifo_type = IfoTypeInfo.find( ( criteria & Query::CRITERIA_BIT_IFO_TYPE )
                                         ? SearchParams.IFOType( )
                                         : SearchParams.IFOTypeAbsolute( ) );

      if ( pos_ifo_type != IfoTypeInfo.end( ) )
      {
        find_time_intervals( SearchParams,
                             Results,
                             pos_ifo_type,
                             Extension );
      }
    }
    else if ( criteria & Query::CRITERIA_BIT_IFO_TYPE_LIST )
    {
      Query     sp( SearchParams );

      for ( Query::ifo_type_list_type::const_iterator
              cur = SearchParams.IFOTypeList( ).begin( ),
              last = SearchParams.IFOTypeList( ).end( );
            cur != last;
            ++cur )
      {
        sp.IFOType( *cur );
        find_ifo_type( sp, Results, IfoTypeInfo, Extension );
      }
    }
    else if ( ( criteria
                & ( Query::CRITERIA_BIT_IFO
                    | Query::CRITERIA_BIT_TYPE ) )
              == ( Query::CRITERIA_BIT_IFO
                   | Query::CRITERIA_BIT_TYPE ) )
    {
      const IfoTypeMatch match( SearchParams.IFO( ),
                                SearchParams.Type( ) );
    
      for( file_base_container_type::const_iterator
             cur = IfoTypeInfo.begin( ),
             last = IfoTypeInfo.end( );
           cur != last;
           ++cur )
      {
        if ( match( cur->first ) )
        {
          find_time_intervals( SearchParams,
                               Results,
                               cur,
                               Extension );
        } // if match
      } // for file_base_container_type
    }
  }
  
  void lHash::
  find_time_intervals( const Query& SearchParams,
                       QueryResults& Results,
                       file_base_container_type::const_iterator IfoTypeInfo,
                       const std::string& Extension ) const
  {
    using diskCache::Query;
    using diskCache::QueryResultsDump;
    using diskCache::QueryResultsFilenames;
    using diskCache::QueryResultsIntervals;
    using diskCache::QueryResultsMountPointStats;
    using diskCache::QueryResultsTimeRange;

    if ( IfoTypeInfo->second == NULL )
    {
      // Bail to prevent SegFault since the pointer is bad
      return;
    }

    qr_intervals*       qr = (qr_intervals*)NULL;
    dir_info_type       di( m_directory_info_ref.lock( ) );

    if ( ( Results.Type( ) == QueryResults::TYPE_INTERNAL )
         && ( ( qr = dynamic_cast< qr_intervals* >( &Results ) ) != NULL )
         && ( di )
         && ( di.get( ) == qr->m_exclude_directory ) )
    {
      //-----------------------------------------------------------------
      // This is the case of collecting interval data and the current
      // directory is the one to be excluded.
      //-----------------------------------------------------------------
      return;
    }
    else if ( ! di )
    {
      //-----------------------------------------------------------------
      // This is the case where the directory information has been
      // released.
      // This hash entry should be removed shortly.
      //-----------------------------------------------------------------
      return;
    }
    TimeHash::KnownIntervals    intervals;

    const std::string path( ( di )
                            ? di->GetPath( ).Var( )
                            : "unknown_dir" );
    const std::string ifo_type( IfoTypeInfo->first );

    //-------------------------------------------------------------------
    // Search for matches within the time hash
    //-------------------------------------------------------------------
    IfoTypeInfo->second->Find( SearchParams, intervals );

    if ( intervals.Ranges( ).empty( ) == false )
    {
      switch( Results.Type( ) )
      {
      case QueryResults::TYPE_DUMP:
        {
          typedef std::map< INT_4U, INT_4U > start_stop_type;
          typedef std::map< INT_4U, start_stop_type > dt_container;

          dt_container  dt_sets;
          for ( TimeHash::KnownIntervals::ranges_type::const_iterator
                  cur_interval = intervals.Ranges( ).begin( ),
                  last_interval = intervals.Ranges( ).end( );
                cur_interval != last_interval;
                ++cur_interval )
          {
            dt_sets[ cur_interval->second.mFileDt ][ cur_interval->first ]
              = cur_interval->second.mStopTime;
          }

          QueryResultsDump&
            r( dynamic_cast< QueryResultsDump& >( Results ) );

          if ( r.Format( ) == QueryResults::FORMAT_TCL )
          {
            static const char* dash = SEPERATOR( );
            static const char* separator = ",";

            const diskCache::mtime_type
              mtime( di->TimeModified( ) );

            std::string::size_type      ifo_pos( 0 );
            std::string::size_type      ifo_len( std::string::npos );
            std::string::size_type      type_pos( ifo_type.find( dash, ifo_pos ) );

            if ( type_pos != std::string::npos )
            {
              ifo_len = type_pos - ifo_pos;
              type_pos++;
            }

            std::string::size_type      type_len( std::string::npos );
            std::string::size_type      end_pos( ifo_type.find( dash, type_pos ) );

            if ( ( type_pos != std::string::npos )
                 && ( end_pos != std::string::npos ) )
            {
              type_len = end_pos - type_pos;
            }

            for ( dt_container::const_iterator
                    cur_dt = dt_sets.begin( ),
                    last_dt = dt_sets.end( );
                  cur_dt != last_dt;
                  ++cur_dt )
            {
              std::ostringstream  result;
            
              INT_4U fc( 0 );

              for ( dt_container::mapped_type::const_iterator
                      cur_time_range = cur_dt->second.begin( ),
                      last_time_range = cur_dt->second.end( );
                    cur_time_range != last_time_range;
                    ++cur_time_range )
              {
                fc += ( ( cur_time_range->second - cur_time_range->first ) / cur_dt->first );
              }

              result
                // Directory path
                << path
                // ifo
                << separator << ifo_type.substr( ifo_pos, ifo_len )
                ;

              result
                // type
                << separator << ifo_type.substr( type_pos, type_len )
                // number of frames per flie (currently hard coded to 1
                << separator << 1
                // delta time
                << separator << cur_dt->first
                // Directory modify time
                << " " << mtime
                // File count
                << " " << fc
                ;
              //-----------------------------------------------------
              // Start/stop gps time pairs of continuous file data
              //-----------------------------------------------------
              r.tcl_list_begin( result );
              bool first = true;
              for ( dt_container::mapped_type::const_iterator
                      cur_time_range = cur_dt->second.begin( ),
                      last_time_range = cur_dt->second.end( );
                    cur_time_range != last_time_range;
                    ++cur_time_range )
              {
                if ( first == false )
                {
                  result << " ";
                }
                else
                {
                  first = false;
                }
                result << cur_time_range->first
                       << " " << cur_time_range->second
                  ;
              }
              result << "}";
              r.Append( result.str( ) );
            }
          }
        }
        break;
      case QueryResults::TYPE_FILENAMES:
        {
          QueryResultsFilenames&
            r( dynamic_cast< QueryResultsFilenames& >( Results ) );

          typedef std::set< std::string >  filename_container;

          std::ostringstream    filename_leader;

          filename_container    filenames;

          filename_leader << path << "/" << ifo_type;

          filename_container::iterator cur_filenames( filenames.begin( ) );

          for ( TimeHash::KnownIntervals::ranges_type::const_iterator
                  cur_interval = intervals.Ranges( ).begin( ),
                  last_interval = intervals.Ranges( ).end( );
                cur_interval != last_interval;
                ++cur_interval )
          {
            Query::gps_time_type start = cur_interval->first;
            Query::gps_time_type last = cur_interval->second.mStopTime;
            const Query::gps_time_type dt = cur_interval->second.mFileDt;

            while ( start < last )
            {
              std::ostringstream      filename;

              filename << filename_leader.str( ) << start << "-" << dt << Extension;
              cur_filenames = filenames.insert( cur_filenames, filename.str( ) );
              start += dt;
            }
          }

          //---------------------------------------------------------
          // Add to the result set
          //---------------------------------------------------------
          for ( filename_container::const_iterator
                  cur_filename = filenames.begin( ),
                  last_filename = filenames.end( );
                cur_filename != last_filename;
                ++cur_filename )
          {
            r.Append( *cur_filename );
          }
        }
        break;
      case QueryResults::TYPE_INTERVALS:
        {
          size_t ifotype_length( ifo_type.length( ) - 1 );
          if ( ifo_type[ ifotype_length ] != '-' )
          {
            ifotype_length = std::string::npos;
          }
          std::ostringstream        key;

          key << ifo_type.substr( 0, ifotype_length )
              << "," << Extension
            ;

          QueryResultsIntervals&
            r( dynamic_cast< QueryResultsIntervals& >( Results ) );

          for ( TimeHash::KnownIntervals::ranges_type::const_iterator
                  cur_interval = intervals.Ranges( ).begin( ),
                  last_interval = intervals.Ranges( ).end( );
                cur_interval != last_interval;
                ++cur_interval )
          {
            // Add the information to the interval results
            r.Append( key.str( ), cur_interval->first, cur_interval->second.mStopTime );
          }
        }
        break;
      case QueryResults::TYPE_MOUNT_POINT_STATS:
        {
          QueryResultsMountPointStats&
            r( dynamic_cast< QueryResultsMountPointStats& >( Results ) );

          r.Append( di->GetMountPointPath( ).Var( ),
                    di->GetPath( ).Var( ),
                    intervals.CountFiles( ) );
        }
        break;
      case QueryResults::TYPE_INTERNAL:
        {
          if ( ( qr )
               && ( ( qr->m_exclude_directory == NULL ) 
                    || ( qr->m_exclude_directory != di.get( ) ) ) )
          {
            //-----------------------------------------------------------
            // This interval should be added to the list as it is
            // not the one being appended
            //-----------------------------------------------------------
            qr->Append( intervals );
          }
        }
        break;
      case QueryResults::TYPE_TIME_RANGE:
        {
          QueryResultsTimeRange&
            r( dynamic_cast< QueryResultsTimeRange& >( Results ) );

          for ( TimeHash::KnownIntervals::ranges_type::const_iterator
                  cur = intervals.Ranges( ).begin( ),
                  last = intervals.Ranges( ).end( );
                cur != last;
                ++cur )
          {
            General::SharedPtr< diskCache::QueryResultsTimeRange::RangeData >
              fd( new FilenameData( ( di )
                                    ? di->GetPath( ).Var( )
                                    : "unknown dir",
                                    ifo_type,
                                    SearchParams.Extension( ),
                                    cur->second.mStopTime,
                                    cur->second.mFileDt,
                                    di->GetMountPointPath( ).Var( ).size( ) ) );
          

            r.Append( cur->first,
                      FilenameData::Reconstitute, fd );
          }
        }
        break;
      default:
        //-----------------------------------------------------------
        /// \todo
        ///     Default case should log warning message into
        ///     system log files.
        //-----------------------------------------------------------
        break;
      }
    }
  }
  
  //-----------------------------------------------------------------
  //
  //-----------------------------------------------------------------
  void qrConflicts::
  Finalize( const Query& SearchParams,
            diskCache::FilenameHandler::IsFileData::conflict_container_type& Conflicts )
  {
    typedef TimeHash::KnownIntervals::ranges_type::mapped_type data_info_type;

    switch( format( ) )
    {
    case FORMAT_CONFLICT:
      //-----------------------------------------------------------------
      // Each of these represents conflicting data
      //-----------------------------------------------------------------
      for ( filename_info_container::const_iterator
	      cur = fileData( ).begin( ),
	      last = fileData( ).end( );
	    cur != last;
	    ++cur )
      {
	const conflict_range existing_data(  cur->first,
					     cur->second.s_data->EndTime( ),
					     cur->second.s_data->DeltaT( ) );
        for ( TimeHash::KnownIntervals::ranges_type::iterator
                nd_cur = m_source.Ranges( ).begin( ),
                nd_last = m_source.Ranges( ).end( );
              nd_cur != nd_last;
              )
        {
	  conflict_range new_data( nd_cur->first,
				   nd_cur->second.mStopTime,
				   nd_cur->second.mFileDt );

	  new_data.intersection( existing_data );
	  if ( new_data.empty( ) == true )
	  {
	    //-----------------------------------------------------------
	    // This is the case where the new data is NOT in conflict.
	    // Advance to the next entry.
	    //-----------------------------------------------------------
	    ++nd_cur;
	    continue;
	  }
          else if ( new_data == existing_data )
          {
            const FilenameData*  fnd( dynamic_cast< const FilenameData* >( cur->second.s_data.get( ) ) );
            if ( m_source_dir->GetPath( ).Var( ).compare( fnd->s_directory ) == 0 )
            {
              //--------------------------------------------------------
              // This is the case where we found ourselves.
              // Advance to the next entry.
              //--------------------------------------------------------
              ++nd_cur;
              continue;
            }
          }

	  //-------------------------------------------------------------
	  // Data has some conflicts:
	  // 1) Generate an appropriate error message
	  //-------------------------------------------------------------
	  conflict_range existing_data_conflict(  existing_data );
	  existing_data_conflict.intersection( new_data );

	  {
	    TimeHash::KnownIntervals::ranges_type::value_type
	      n( new_data.m_start,
		 data_info_type( new_data.m_end, new_data.m_dt ) );

	    m_conflicts.Ranges( ).insert( n );
	  }
          const FilenameData*  fnd( dynamic_cast< const FilenameData* >( cur->second.s_data.get( ) ) );
	  std::ostringstream	conflict_msg;

          std::ostringstream    ifo_type;

          if ( fnd )
          {
            ifo_type << "IFO/Type: " << fnd->s_ifo_type
                     << " Extension: " << fnd->s_extension
                     << " "
              ;
          }
	  
	  conflict_msg
            << "Time overlap is detected for frame interval " << new_data
            << " in directory " << m_source_dir->GetPath( ).Var( )
	    << " and already existing frame interval " << existing_data_conflict
	    << " in directory " << fnd->s_directory
	    << " ("
            << ifo_type.str( )
	    << ( ( m_source_dir->GetMountPointPath( ).Var( )
                   ==  fnd->s_directory.substr( 0, fnd->s_mount_point ) )
                 ? "SAME" : "DIFFERENT" )
	    << " MOUNT POINT)."
	    ;
	    

          Conflicts.push_back( conflict_msg.str( ) );

	  //-------------------------------------------------------------
	  // 2) Remove the conflicting data from the list
	  //-------------------------------------------------------------
	  if ( ( new_data.m_start == nd_cur->first )
	       && ( new_data.m_end == nd_cur->second.mStopTime ) )
	  {
	    // Everything was in conflict
	    TimeHash::KnownIntervals::ranges_type::iterator
	      conflict_iterator( nd_cur );

	    ++nd_cur;
	    m_source.Ranges( ).erase( conflict_iterator );
	  }
	  else if ( ( new_data.m_start > nd_cur->first )
		    && ( new_data.m_end == nd_cur->second.mStopTime ) )
	  {
	    //-----------------------------------------------------------
	    // Preserve the front
	    //-----------------------------------------------------------
	    nd_cur->second.mStopTime = new_data.m_start;
	    ++nd_cur;
	  }
	  else if ( ( new_data.m_start == nd_cur->first )
		    && ( new_data.m_end < nd_cur->second.mStopTime ) )
	  {
	    //-----------------------------------------------------------
	    // Preserve the back
	    //-----------------------------------------------------------
	    TimeHash::KnownIntervals::ranges_type::value_type
	      n( new_data.m_end,
		 data_info_type( nd_cur->second.mStopTime, new_data.m_dt ) );

	    m_source.Ranges( ).erase( nd_cur );
	    nd_cur = m_source.Ranges( ).insert( n ).first;
	  }
	  else
	  {
	    //-----------------------------------------------------------
	    // In this case the conflict is in the middle
	    //-----------------------------------------------------------

	    //-----------------------------------------------------------
	    // Preserve the back
	    //-----------------------------------------------------------
	    TimeHash::KnownIntervals::ranges_type::value_type
	      n( new_data.m_end,
		 data_info_type( nd_cur->second.mStopTime, new_data.m_dt ) );

	    //-----------------------------------------------------------
	    // Preserve the front
	    //-----------------------------------------------------------
	    nd_cur->second.mStopTime = new_data.m_start;

	    //-----------------------------------------------------------
	    // Put in the back section and continue validation at
	    // that point
	    //-----------------------------------------------------------
	    nd_cur = m_source.Ranges( ).insert( n ).first;
	  }

	}
      }
      break;
    default:
      break;
    }
  } /* Finalize */
} /* namespace - anonymous */
