/* -*- c-basic-offset: 3; indent-tabs-mode: nil -*- */
#ifndef DISKCACHE_CACHE_FILE_HH
#define DISKCACHE_CACHE_FILE_HH
   
   
// System Header Files   
#include <list>   
#include <map>   
#include <set>
#include <string>   
#include <typeinfo>
#include <vector>   
   
// General Header Files   
#include "general/autoarray.hh"
#include "general/fstream.hh"
#include "general/reverse.hh"
#include "general/SharedPtr.hh"
#include "general/types.hh"

#include "genericAPI/Logging.hh"

// Local Header Files   
#include "rwsingleton.hh"

#define DISKCACHE__CACHE_FILE_VERBOSE_DEBUGGING 0
   
/*! \file
 *
 *  The declaration of diskCache::CacheFile.
 */
namespace diskCache
{
   ///\brief Frame cache file (singleton) class.
   ///
   /// This class is designed to read and write frame cache from/to
   /// the binary file (it's assumed that reads and writes will be done on the same 
   /// platform).
   ///
   /// This singleton class is not tied up to a particular file name: Tcl layer is
   /// allowed to specify different files for read and write operations. The only 
   /// restriction is that only one read or write can happen at a time (singleton's
   /// purpose).
   ///
   class CacheFile
   {
   public: 
      typedef INT_8U size_type;
      typedef INT_2U version_type;
      typedef ::General::fstream        stream_type;

      static const version_type CURRENT_VERSION;

      enum mode_type {
         CF_MODE_BINARY,
         CF_MODE_TCL
      };

      //-----------------------------------------------------------------
      /// \brief Validate the version number
      //-----------------------------------------------------------------
      static void ValidateVersion( version_type Version );

      // Accessor
      /// \brief View the next available input character.
      std::ios::int_type peek();

      /// \brief Write current frame hash to the binary file.
      void writeToFile( const CHAR* const tcl_file_name,
                        mode_type Mode = CF_MODE_BINARY,
                        version_type Version = CURRENT_VERSION );

      // Reader (blocking)
      /// \brief Read binary cache file from file.
      void readFromFile( const CHAR* const tcl_file_name );
   
      // These overloaded >> operators are not locked: know what you're doing!
      // They are public to avoid too many friends that need to use them.

#if PURGE
      template < typename T >
      inline CacheFile& operator<<( const General::SharedPtr< T > Value )
      {
         return operator<<( *Value );
      }
#endif /* PURGE */

      template < typename T >
      CacheFile& operator<<( const T& Value );

#if PURGE
      template < typename T >
      inline CacheFile& operator>>( General::SharedPtr< T > Value )
      {
         return operator>>( *Value );
      }
#endif /* PURGE */
   
      template < typename T >
      CacheFile& operator>>( T& Value );
   
      version_type Version( ) const;

   private:
      /// \brief Unformatted read from the file.
      void read( void* ptr, const std::streamsize& size );

      template < typename Container >
      CacheFile& read_container( Container& Data );

      template < typename Element >
      CacheFile& read_element( Element& Data );

      /// \brief Unformatted write to the file.
      void write( const void* ptr, const std::streamsize& size );
   
      template < typename Container >
      CacheFile& write_container( const Container& Data );

      template < typename Element >
      CacheFile& write_element( const Element& Data );

      /// Stream attached to the file.
      stream_type	                mStream;

      General::AutoArray< char >        m_stream_buffer;
      /// Flag indicating the need to byte swap the data.
      bool      	m_need_to_swap_data;

      version_type      m_version;
   };
   

   inline CacheFile::version_type CacheFile::
   Version( ) const
   {
      return m_version;
   }

   template< typename Element >
   inline CacheFile& CacheFile::
   read_element( Element& Data )
   {
      using namespace General;

      static const char caller[] = "diskCache::CacheFile::read_element";

      read( &Data, sizeof( Element ) );

      if ( m_need_to_swap_data )
      {
         reverse< sizeof( Element ) >( &Data, 1 );
      }

      if ( DISKCACHE__CACHE_FILE_VERBOSE_DEBUGGING )
      {
         using GenericAPI::LogEntryGroup_type;
         using GenericAPI::queueLogEntry;

         std::ostringstream        msg;

         msg << "Read element of size: " << sizeof( Element )
             << " pos: " << (INT_8U)( mStream.tellg( ) )
             << " Data: " << Data
             << " Data (hex): " << std::hex << Data
            ;
         queueLogEntry( msg.str( ),
                        LogEntryGroup_type::MT_DEBUG,
                        40,
                        caller,
                        "CXX_INTERFACE" );
      }

      return *this;
   }

   template<>
   inline CacheFile& CacheFile::
   operator>>( INT_2S& Data )
   {
      return read_element( Data );
   }

   template<>
   inline CacheFile& CacheFile::
   operator>>( INT_2U& Data )
   {
      return read_element( Data );
   }

   template<>
   inline CacheFile& CacheFile::
   operator>>( INT_4S& Data )
   {
      return read_element( Data );
   }

   template<>
   inline CacheFile& CacheFile::
   operator>>( INT_4U& Data )
   {
      return read_element( Data );
   }

   template<>
   inline CacheFile& CacheFile::
   operator>>( INT_8S& Data )
   {
      return read_element( Data );
   }

   template<>
   inline CacheFile& CacheFile::
   operator>>( INT_8U& Data )
   {
      return read_element( Data );
   }

   template<>
   inline CacheFile& CacheFile::
   operator>>( std::string& Data )
   {
      static const char caller[] = "diskCache::CacheFile::operator>>( std::string& )";
      using namespace General;

      Data.erase( );

      typedef INT_8U string_size_type;

      string_size_type  string_size( 0 );

      (*this) >> string_size;

      if ( string_size > 0 )
      {
         AutoArray< CHAR > buffer( new CHAR[ string_size + 1 ] );
         
         read( buffer.get( ), sizeof( CHAR ) * string_size );
         Data.assign( buffer.get( ), string_size );
         if ( DISKCACHE__CACHE_FILE_VERBOSE_DEBUGGING )
         {
            using GenericAPI::LogEntryGroup_type;
            using GenericAPI::queueLogEntry;

            std::ostringstream        msg;

            msg << "Read string: "
                << " Data: " << Data
               ;
            queueLogEntry( msg.str( ),
                           LogEntryGroup_type::MT_DEBUG,
                           40,
                           caller,
                           "CXX_INTERFACE" );
         }
      }
   
      return *this;
   }

   template < typename Container >
   inline CacheFile& CacheFile::
   read_container( Container& Data )
   {
      static const char caller[] = "diskCache::CacheFile::read_container";
      INT_8U container_size( 0 );

      (*this) >> container_size;

      if ( DISKCACHE__CACHE_FILE_VERBOSE_DEBUGGING )
      {
         using GenericAPI::LogEntryGroup_type;
         using GenericAPI::queueLogEntry;

         std::ostringstream        msg;

         msg << "This container has: " << container_size
             << " element" << ( ( container_size == 1 ) ? "" : "s" )
             << " of type: " << typeid( Data ).name( )
            ;
         queueLogEntry( msg.str( ),
                        LogEntryGroup_type::MT_DEBUG,
                        40,
                        caller,
                        "CXX_INTERFACE" );
      }
      if ( container_size > 0 )
      {
         Data.resize( container_size );

         for ( typename Container::iterator
                  cur = Data.begin( ),
                  last = Data.end( );
               cur != last;
               ++cur )
         {
            (*this) >> (*cur);
         }
      }
      else
      {
         Data.erase( Data.begin( ), Data.end( ) );
      }
      return *this;
   }

   template <>
   inline CacheFile& CacheFile::
   operator>>( std::list< std::string >& Data )
   {
      return read_container( Data );
   }

   template <>
   inline CacheFile& CacheFile::
   operator>>( std::set< std::string >& Data )
   {
      typedef std::set< std::string >   data_type;

      static const char caller[] = "diskCache::CacheFile::read_container";
      INT_8U container_size( 0 );

      (*this) >> container_size;

      if ( DISKCACHE__CACHE_FILE_VERBOSE_DEBUGGING )
      {
         using GenericAPI::LogEntryGroup_type;
         using GenericAPI::queueLogEntry;

         std::ostringstream        msg;

         msg << "This container has: " << container_size
             << " element" << ( ( container_size == 1 ) ? "" : "s" )
             << " of type: " << typeid( Data ).name( )
            ;
         queueLogEntry( msg.str( ),
                        LogEntryGroup_type::MT_DEBUG,
                        40,
                        caller,
                        "CXX_INTERFACE" );
      }

      for ( INT_8U
               x = 0;
            x < container_size;
            ++x )
      {
         data_type::value_type       v;

         (*this) >> v;
         Data.insert( v );
      }
      return *this;
   }

   template <>
   inline CacheFile& CacheFile::
   operator>>( std::vector< std::string >& Data )
   {
      return read_container( Data );
   }

   template< typename Element >
   inline CacheFile& CacheFile::
   write_element( const Element& Data )
   {
      using namespace General;

      static const char caller[] = "diskCache::CacheFile::write_element";

      write( &Data, sizeof( Element ) );

      if ( DISKCACHE__CACHE_FILE_VERBOSE_DEBUGGING )
      {
         using GenericAPI::LogEntryGroup_type;
         using GenericAPI::queueLogEntry;

         std::ostringstream        msg;

         msg << "Write element of size: " << sizeof( Element )
             << " pos: " << (INT_8U)( mStream.tellg( ) )
             << " Data: " << Data
             << " Data (hex): " << std::hex << Data
            ;
         queueLogEntry( msg.str( ),
                        LogEntryGroup_type::MT_DEBUG,
                        40,
                        caller,
                        "CXX_INTERFACE" );
      }

      return *this;
   }

   template<>
   inline CacheFile& CacheFile::
   operator<<( const INT_2S& Data )
   {
      return write_element( Data );
   }

   template<>
   inline CacheFile& CacheFile::
   operator<<( const INT_2U& Data )
   {
      return write_element( Data );
   }

   template<>
   inline CacheFile& CacheFile::
   operator<<( const INT_4S& Data )
   {
      return write_element( Data );
   }

   template<>
   inline CacheFile& CacheFile::
   operator<<( const INT_4U& Data )
   {
      return write_element( Data );
   }

   template<>
   inline CacheFile& CacheFile::
   operator<<( const INT_8S& Data )
   {
      return write_element( Data );
   }

   template<>
   inline CacheFile& CacheFile::
   operator<<( const INT_8U& Data )
   {
      return write_element( Data );
   }

   template<>
   inline CacheFile& CacheFile::
   operator<<( const std::string& Data )
   {
      static const char caller[] = "diskCache::CacheFile::operator<<( const std::string& )";
      using namespace General;

      typedef INT_8U string_size_type;

      (*this) << ( INT_8U )( Data.length( ) );

      if ( Data.length( ) > 0 )
      {
         write( Data.c_str( ), sizeof( CHAR ) * Data.length( ) );

         if ( DISKCACHE__CACHE_FILE_VERBOSE_DEBUGGING )
         {
            using GenericAPI::LogEntryGroup_type;
            using GenericAPI::queueLogEntry;

            std::ostringstream        msg;

            msg << "Wrote string: "
                << " Data: " << Data
               ;
            queueLogEntry( msg.str( ),
                           LogEntryGroup_type::MT_DEBUG,
                           40,
                           caller,
                           "CXX_INTERFACE" );
         }
      }
   
      return *this;
   }

   template < typename Container >
   inline CacheFile& CacheFile::
   write_container( const Container& Data )
   {
      static const char caller[] = "diskCache::CacheFile::write_container";
      INT_8U container_size( Data.size( ) );

      (*this) << container_size;

      if ( DISKCACHE__CACHE_FILE_VERBOSE_DEBUGGING )
      {
         using GenericAPI::LogEntryGroup_type;
         using GenericAPI::queueLogEntry;

         std::ostringstream        msg;

         msg << "This container has: " << container_size
             << " element" << ( ( container_size == 1 ) ? "" : "s" )
             << " of type: " << typeid( Data ).name( )
            ;
         queueLogEntry( msg.str( ),
                        LogEntryGroup_type::MT_DEBUG,
                        40,
                        caller,
                        "CXX_INTERFACE" );
      }
      if ( container_size > 0 )
      {
         for ( typename Container::const_iterator
                  cur = Data.begin( ),
                  last = Data.end( );
               cur != last;
               ++cur )
         {
            (*this) << (*cur);
         }
      }
      return *this;
   }


   template <>
   inline CacheFile& CacheFile::
   operator<<( const std::list< std::string >& Data )
   {
      return write_container( Data );
   }

   template <>
   inline CacheFile& CacheFile::
   operator<<( const std::set< std::string >& Data )
   {
      return write_container( Data );
   }

   template <>
   inline CacheFile& CacheFile::
   operator<<( const std::vector< std::string >& Data )
   {
      return write_container( Data );
   }

}
   
   
#endif   
