/* -*- mode: c++; c-basic-offset: 3; indent-tabs-mode: nil; -*- */

#ifndef DISKCACHE_DIRECTORY_INFO_HH
#define DISKCACHE_DIRECTORY_INFO_HH

#include <iostream>
   
// System Header Files   
#include <list>   
#include <set>
#include <stdexcept>
#include <string>   
#include <utility>   
#include <vector>   
   
// General Header Files   
#include "general/gpstime.hh"
#include "general/ReadWriteLock.hh"
#include "general/SharedPtr.hh"
#include "general/types.hh"

// Local Header Files   
#include "diskcacheAPI.hh"
#include "timehash.hh"   // for KnownIntervals
#include "FilenameHandler.hh"
#include "mountpt.hh"
   
namespace diskCache
{
   // Forward declarations
   class CacheFile;
   class Query;
   class QueryResults;
   class LogMessage;
   
   //: Directory information (thread safe) class.
   //
   // This class is designed to keep all necessary information about
   // a particular directory.
   //
   class DirectoryInfo
   {
   public:
      //-----------------------------------------------------------------
      /// \brief A resulting collection of directories
      ///
      /// This is a collection of directories being returned
      /// from a method.
      /// This list is not necessarily ordered.
      //-----------------------------------------------------------------
      typedef MountPointHash::resulting_dirs_container_type
      resulting_dirs_container_type;

      typedef MountPointHash::excluded_dirs_container_type
      excluded_dirs_container_type;

#if 0
      typedef std::list< std::string >          dir_container_type;
#else
      typedef std::set< std::string >           dir_container_type;
#endif /* 0 */
      typedef General::ReadWriteLockVariable< General::ReadWriteLock::READ,
                                              std::string >
      path_variable_ro_type;
      typedef FilenameHandler::HashContainer    index_type;

      typedef General::ReadWriteLockVariable< General::ReadWriteLock::READ,
                                              index_type >
      index_variable_ro_type;
      typedef General::ReadWriteLockVariable< General::ReadWriteLock::WRITE,
                                              index_type >
      index_variable_rw_type;
      
      // Constructors
      DirectoryInfo( const std::string& path, const std::string& mount_pt );
   
      // Destructor
      ~DirectoryInfo();

   
      //-----------------------------------------------------------------
      /// \brief Modify the subdirectory metadata.
      ///
      /// \param[in] AddedToList
      ///     An ordered container of directories that have been
      ///     added to the list directories to be excluded.
      /// \param[in] RemovedFromList
      ///     An ordered container of directories that have been
      ///     removed from the list of directories to be excluded.
      /// \param[out] RemovedDirectories
      ///     A list of removed subdirectories with frame data.
      ///     This list will be empty if no data was removed as a
      ///     result of directory removal.
      ///     Returned list will include parent directory itself if it
      ///     contains frame data.       
      /// \param[out] Errors
      ///     Accumulative error message. Appended messsage must be
      ///     enclosed in Tcl " {} " list.
      //-----------------------------------------------------------------
      void ExcludedDirs( const excluded_dirs_container_type& AddedToList,
                         const excluded_dirs_container_type& RemovedFromList,
                         resulting_dirs_container_type& RemovedDirectories,
                         LogMessage& Errors );

      //-----------------------------------------------------------------
      /// \brief Retrieve the number of files being managed
      ///
      /// \return
      ///     The number of files being managed under this directory.
      //-----------------------------------------------------------------
      INT_4U FileCount( ) const;

      void setExcludedSubDir( const std::string& sdir );
   
      //-----------------------------------------------------------------
      /// \brief Removes all known subdirectories from global hash.
      ///
      /// \param[out] RemovedDirectories
      ///     A list of removed subdirectories with frame data.
      ///     This list will be empty if no data was removed as a
      ///     result of directory removal.
      ///     Returned list will include parent directory itself if it
      ///     contains frame data.       
      /// \param[out] errors
      ///     Accumulative error message. Appended messsage must be
      ///     enclosed in Tcl " {} " list.
      /// \param[in] unregister
      ///     A flag to indicate if directory needs to  unregister
      ///     itself with global frame hash. Set to FALSE only when
      ///     MOUNT_PT directory is removed ===> the whole entry 
      ///     for MOUNT_PT directory is removed from frame hash at once.
      //-----------------------------------------------------------------
      void removeSubDirectories( resulting_dirs_container_type& RemovedDirectories,
                                 LogMessage& errors,
                                 const bool unregister );   

   
      //-----------------------------------------------------------------
      /// \brief Scan directory recursively
      /// 
      /// \param[in] exclude_dirs
      ///     Patterns of directories to exclude from being searched.
      /// \param[out] errors
      ///     Accumulative error message. Appended messsage must be
      ///     enclosed in Tcl " {} " list.
      /// \param[out] Modified
      ///     True if the directory node has been replaced with a new
      ///     node; false otherwise.
      //-----------------------------------------------------------------
      std::list< std::string >
      scan( const excluded_dirs_container_type& exclude_dirs,
            LogMessage& errors,
            bool& Modified );
   
      // Accessors
      const int getHashNumbers( const std::string& ifo_type = "",
                                bool TryLock = false ) const;

      // FIX PR 2390: new query method: pass start,stop into search
      const TimeHash::KnownIntervals getFrameQueryIntervals( 
                                              const std::string& ifo_type,
                                              const INT_4U query_start, 
                                              const INT_4U query_stop ) const;

      const std::list< std::string > getSubDirectoryList() const;   

      const std::string getPath( bool TryLock = false ) const;
   
      const path_variable_ro_type GetPath( ) const;

      void GetSubDirectories( std::list< std::string >& Var,
                              std::list< std::string >::iterator Start ) const;
   
      const bool isMountPtEntry() const;   
   
      const std::string getMountPtPath() const;
   
      const path_variable_ro_type GetMountPointPath( ) const;

      const TimeHash::KnownIntervals getKnownIntervals( 
                                       const std::string& ifo_type ) const;
      
      bool HasData( ) const;

      void Find( const Query& SearchParams,
                 QueryResults& Results ) const;

      //-----------------------------------------------------------------
      /// \brief Retrieve a reference to the mount point path
      ///
      /// \return
      ///     A reference to the storage containing the mount ponit
      ///     path associated with this object.
      ///
      /// \note
      ///     This should only be called by functions that have locked
      ///     this object at some other level.
      //-----------------------------------------------------------------
      const std::string& RefMountPointPath( ) const;

   
      //: Writer
      void write( CacheFile& stream ) const;
   
      //: Reader
      static General::SharedPtr< DirectoryInfo > read( CacheFile& stream );

   
      mtime_type TimeModified( ) const;

      void TimeModified( mtime_type Value );

      void TimeModifiedReset( );

      index_variable_rw_type Index( );

      const index_variable_ro_type Index( ) const;

      const index_variable_ro_type IndexRO( ) const;

   private:
      enum {
         MODIFIED_RESET,
         MODIFIED_NEW_DIRECTORY
      };
      //: No default constructor
      DirectoryInfo();
   
      //: No copy constructor
      DirectoryInfo( const DirectoryInfo& );   
   
      //: No operator=
      const DirectoryInfo& operator=( const DirectoryInfo& );      
   
      //-----------------------------------------------------------------
      /// \brief Read object's internal data from the stream.
      ///
      /// \param[in] stream
      ///     The stream from which to reconstruct the object.
      /// \param[in] DirInfo
      ///     Reference to self
      //-----------------------------------------------------------------
      void readInfo( CacheFile& stream,
                     General::SharedPtr< DirectoryInfo > DirInfo );
   
   
      //-----------------------------------------------------------------
      /// \brief Prevent multiple scans of this object
      //-----------------------------------------------------------------
      mutable MutexLock::lock_type m_scan_baton;

      //: Object lock
      mutable General::ReadWriteLock::lock_type mLock;

      //-----------------------------------------------------------------
      /// \brief Object lock for constant data
      //-----------------------------------------------------------------
      mutable General::ReadWriteLock::lock_type m_const_lock;

      //: Directory's own full path
      const std::string mPath;
   
      //: Mount point under which directory resides.
      const std::string mMountPtPath;

      //: Sorted list of all sub-directories 
      //: (they are entries in global DirectoryHash)
      dir_container_type mSubDirs;      
   
      //-----------------------------------------------------------------
      /// \brief Sorted list of all excluded sub-directories 
      ///
      /// This list contains only the base name of the directory
      ///   matching one entry of MountPointHash::mExcludeDirList.
      //-----------------------------------------------------------------
      dir_container_type mExcludedSubDirs;         
   
      //-----------------------------------------------------------------
      /// \brief Count of regular files in the local cache
      //-----------------------------------------------------------------
      mutable INT_4U    m_file_count;

      //-----------------------------------------------------------------
      /// \brief Last time that the directory was modified
      //-----------------------------------------------------------------
      mtime_type        m_last_modified;

      //-----------------------------------------------------------------
      /// \brief Collection of indexed files
      //-----------------------------------------------------------------
      index_type                                p_index;
      mutable General::ReadWriteLock::lock_type p_index_baton;
   
      index_variable_rw_type get_index( );

      const index_variable_ro_type get_index( ) const;

      //-----------------------------------------------------------------
      /// \brief Check if any regular files are being managed.
      //-----------------------------------------------------------------
      bool has_data( ) const;

      const std::string& mount_point( ) const;

      //-----------------------------------------------------------------
      /// \brief Read the directory contents.
      ///
      /// \param[in] ExcludedDirectories
      ///     A collection of directory patterns that are not to be
      ///     analysed.
      /// \param[out] errors
      ///     A collection of errors that have been accumulated during
      ///     the current scan.
      /// \param[in] PreviousInfo
      ///     A reference to the previous contents of the directory.
      /// \param[in,out] CurrentInfo
      ///     Storage for the current contents of the directory.
      /// \param[out] AddedFileCount
      ///     The number of files that have been added relative to
      ///     the previously cached information about the directory.
      /// \param[out] RemovedFileCount
      ///     The number of files that have been removed relative to
      ///     the previously cached information about the directory.
      //-----------------------------------------------------------------
      void read_directory( const excluded_dirs_container_type& ExcludedDirectories,
                           LogMessage& errors,
                           const DirectoryInfo& PreviousInfo,
                           General::SharedPtr< DirectoryInfo > CurrentInfo,
                           INT_4U& AddedFileCount,
                           INT_4U& RemovedFileCount );
   };

   inline INT_4U DirectoryInfo::
   FileCount( ) const
   {
      return m_file_count;
   }

   inline void DirectoryInfo::
   GetSubDirectories( std::list< std::string >& Var,
                      std::list< std::string >::iterator Start ) const
   {
      General::ReadWriteLock lock( mLock, General::ReadWriteLock::READ,
                                   __FILE__, __LINE__ );

#if WORKING
      std::string path( mPath );
      path += "/";
      std::string::size_type    offset = path.length( );
#endif /* WORKING */

      std::list< std::string > paths;
      for( dir_container_type::const_iterator
              cur = mSubDirs.begin( ),
              last = mSubDirs.end( );
           cur != last;
           ++cur )
      {
#if WORKING
         path.replace( offset, std::string::npos, *cur );
         Start = Var.insert( Start, path );
#else /* WORKING */
         Start = Var.insert( Start, *cur );
#endif /* WORKING */
         Start++;
      }

      Var.insert( Start, paths.begin( ), paths.end( ) );
   }
   
   inline const DirectoryInfo::path_variable_ro_type DirectoryInfo::
   GetPath( ) const
   {
      return( path_variable_ro_type( m_const_lock,
                                     RWLOCK_TIMEOUT,
                                     mPath,
                                     __FILE__, __LINE__ ) )
         ;
   }
   
   inline const DirectoryInfo::path_variable_ro_type DirectoryInfo::
   GetMountPointPath( ) const
   {
      return( path_variable_ro_type( m_const_lock,
                                     RWLOCK_TIMEOUT,
                                     mMountPtPath,
                                     __FILE__, __LINE__ ) )
         ;
   }

   inline const std::string& DirectoryInfo::
   RefMountPointPath( ) const
   {
      return mount_point( );
   }

   inline mtime_type DirectoryInfo::
   TimeModified( ) const
   {
      //---------------------------------------------------------
      // :TODO: Protect with locks
      //---------------------------------------------------------
      return m_last_modified;
   }

   inline void DirectoryInfo::
   TimeModified( mtime_type Value )
   {
      //---------------------------------------------------------
      // :TODO: Protect with locks
      //---------------------------------------------------------
      //---------------------------------------------------------
      // Protect the setting of the modified time from
      // the 1 second bug by setting to 0 m_last_modified
      // if Value is within 1 second of the the current
      // time.
      //---------------------------------------------------------
      using General::GPSTime;

      GPSTime   dir( Value, 0, GPSTime::UTC );
      GPSTime   now;
      now.Now( );

      if ( ( now - dir ) <= 1.0 )
      {
         m_last_modified = MODIFIED_RESET;
      }
      else
      {
         m_last_modified = Value;
      }
   }

   inline void DirectoryInfo::
   TimeModifiedReset( )
   {
      //---------------------------------------------------------
      // :TODO: Protect with locks
      //---------------------------------------------------------
      m_last_modified = MODIFIED_RESET;
   }

   inline bool DirectoryInfo::
   HasData( ) const
   {
      return has_data( );
   }

   inline DirectoryInfo::index_variable_rw_type DirectoryInfo::
   Index( )
   {
      return get_index( );
   }
 
   inline const DirectoryInfo::index_variable_ro_type DirectoryInfo::
   Index( ) const
   {
      return get_index( );
   }
 
   inline const DirectoryInfo::index_variable_ro_type DirectoryInfo::
   IndexRO( ) const
   {
      return get_index( );
   }
 
   inline DirectoryInfo::index_variable_rw_type DirectoryInfo::
   get_index( )
   {
      return( index_variable_rw_type( p_index_baton,
                                      0,
                                      p_index,
                                      __FILE__, __LINE__ ) );
   }
 
   inline const DirectoryInfo::index_variable_ro_type DirectoryInfo::
   get_index( ) const
   {
      return( index_variable_ro_type( p_index_baton,
                                      RWLOCK_TIMEOUT,
                                      p_index,
                                      __FILE__, __LINE__ ) );
   }
 
   inline bool DirectoryInfo::
   has_data( ) const
   {
      return ( m_file_count> 0 );
   }

   inline const std::string& DirectoryInfo::
   mount_point( ) const
   {
      return mMountPtPath;
   }
}
   
   
#endif
