/* -*- mode: c++; c-basic-offset: 3; indent-tabs-mode: nil; -*- */

#ifndef DISKCACHE_MOUNT_PT_HASH_HH
#define DISKCACHE_MOUNT_PT_HASH_HH

// System Header Files   
#include <sys/statvfs.h>

#include <list>   
#include <set>
#include <stdexcept>
#include <string>
#include <vector>   
   
// General Header Files   
#include "general/types.hh"

#include "general/unordered_map.hh"
#include "general/ReadWriteLock.hh"
#include "general/SharedPtr.hh"
      
// Local Header Files   
#include "diskcacheAPI.hh"
#include "timehash.hh"   // KnownIntervals structure
#include "MountPointCache.hh"
   
   
namespace diskCache
{
   //--------------------------------------------------------------------
   // Forward class declarations
   //--------------------------------------------------------------------

   class CacheFile;
   class DirectoryInfo;
   class LogMessage;
   class Query;
   class QueryResults;

   //--------------------------------------------------------------------
   /// \brief Mount point hash (singleton) class.
   ///
   /// This class is designed to keep track of all directory entries
   /// that appear only in MOUNT_PT list (diskcacheAPI resource
   /// file variable).
   ///
   //--------------------------------------------------------------------
   class MountPointHash
   {
   public: 
      class Conflict
         : public std::runtime_error
      {
      public:
         Conflict( const std::string& Message );
      };

      //-----------------------------------------------------------------
      /// \brief A resulting collection of directories
      ///
      /// This is a collection of directories being returned
      /// from a method.
      /// This list is not necessarily ordered.
      //-----------------------------------------------------------------
      typedef std::list< std::string > resulting_dirs_container_type;
      //-----------------------------------------------------------------
      /// \brief An ordered collection of directory names being excluded
      ///
      /// This is a collection of directories that should be excluded
      /// from discovery.
      /// It is ordered to enhance performance.
      //-----------------------------------------------------------------
      typedef std::set< std::string > excluded_dirs_container_type;

      typedef General::ReadWriteLockVariable< General::ReadWriteLock::READ,
                                              excluded_dirs_container_type >
      excluded_dirs_ro_type;
      typedef General::ReadWriteLockVariable< General::ReadWriteLock::WRITE,
                                              excluded_dirs_container_type >
      excluded_dirs_rw_type;

      //-----------------------------------------------------------------
      //-----------------------------------------------------------------
      typedef std::vector< std::string > mount_point_container_type;

      typedef General::ReadWriteLockVariable< General::ReadWriteLock::READ,
                                           mount_point_container_type >
      mount_point_ro_type;
     typedef General::ReadWriteLockVariable< General::ReadWriteLock::WRITE,
                                           mount_point_container_type >
      mount_point_rw_type;
                                           
      typedef General::SharedPtr< DirectoryInfo > dir_info_type;

      //-----------------------------------------------------------------
      //-----------------------------------------------------------------
      static const INT_4U               SCANNED_MOUNT_POINTS_RESET;

      static void Register( dir_info_type DirInfo );

      //-----------------------------------------------------------------
      // \brief Replace one node with another.
      //-----------------------------------------------------------------
      static void Replace( dir_info_type DirInfoOld,
                           dir_info_type DirInfoNew );

      //-----------------------------------------------------------------
      /// \brief Reset the mount point hash to its initial state
      //-----------------------------------------------------------------
      static void Reset( );

      static void Unregister( dir_info_type DirInfo );

      static bool CheckForDataConflictsAcrossAllMountPoints( );

      //-----------------------------------------------------------------
      /// \brief Search for information within the cache
      ///
      /// \param[in] SearchParams
      ///     The search parameters describing the data to find.
      /// \param[out] Results
      ///     The data found.
      //-----------------------------------------------------------------
      static void Find( const Query& SearchParams,
                        QueryResults& Results );

      inline static General::GPSTime LastUpdate( )
      {
         MutexLock      lock( m_last_update_baton );

         return m_last_update;
      }

      inline static void LastUpdateSet( )
      {
         MutexLock      lock( m_last_update_baton );

         m_last_update.Now( );
      }

      static const excluded_dirs_ro_type ExcludedDirs( );

      static const mount_point_ro_type MountPoints( const char* File = __FILE__,
                                                    int Line = __LINE__ );

      inline static INT_4U ScannedMountPoints( )
      {
         MutexLock      lock( m_scanned_mount_points_baton );

         return m_scanned_mount_points;
      }

      inline static void ScannedMountPoints( INT_4U Value )
      {
         MutexLock      lock( m_scanned_mount_points_baton );

         if ( Value == SCANNED_MOUNT_POINTS_RESET )
         {
            m_scanned_mount_points = 0;
         }
         else
         {
            m_scanned_mount_points += Value;
         }
      }

      //-----------------------------------------------------------------
      /// \brief Initialize/update MOUNT_PT list
      ///
      /// \attention
      ///     Please place VERY gappy data at the end of the MOUNT_PT list!
      ///     It greatly affects time spent in data look ups (satisfying 
      ///     frame queries): ~ 50-90 times slower to copy and step through 
      ///     very gappy data.
      ///
      /// \param[in] dir_list
      ///     A list of directories as they appear in MOUNT_PT list
      ///     (API resource file variable).
      /// \param[in] enable_global_check
      ///     A flag to indicate if API should check for data conflicts
      ///     across all MOUNT_PT entries (diskcacheAPI\@LDAS reports
      ///     data conflicts only under the same MOUNT_PT).      
      ///   
      /// \return
      ///     Sorted Tcl list of all removed subdirectories, followed by
      ///     error messages if any: { subdirs } { {error1} {error2} }
      //-----------------------------------------------------------------
      static const std::string updateMountPtList( const CHAR* const dir_list,
                                                  const bool enable_global_check );
   
      //-----------------------------------------------------------------
      /// \brief Initialize/update MOUNT_PT list
      ///
      /// \attention
      ///     Please place VERY gappy data at the end of the MOUNT_PT list!
      ///     It greatly affects time spent in data look ups (satisfying 
      ///     frame queries): ~ 50-90 times slower to copy and step through 
      ///     very gappy data.
      ///
      /// \param[in,out] MountPointCollection
      ///     A list of directories as they appear in MOUNT_PT list
      ///     (API resource file variable).
      /// \param[out] Added
      ///     A collection of mount point directories that have been added.
      /// \param[out] Removed
      ///     A collection of mount point directories that have been removed.
      /// \param[out] SubDirsRemoved
      ///     A collection of directories that have been removed
      ///     from under the mount points.
      /// \param[out] Errors
      ///     A list of errors encountered when trying to remove \a dir.
      ///   
      //-----------------------------------------------------------------
      static void
      UpdateMountPointList( mount_point_container_type& NewMountPoints,
                            mount_point_container_type& Added,
                            mount_point_container_type& Removed,
                            std::list< mount_point_container_type::value_type >&
                            SubDirsRemoved,
                            LogMessage& Errors );

      //-----------------------------------------------------------------
      /// \brief  Update excluded directory list
      ///
      /// \param dir_list
      ///      A list of directories that should be excluded from frame
      ///      data hash.
      ///
      /// \return
      ///     Sorted Tcl list of all removed subdirectories, followed
      ///     by error messages if any: { subdirs }{ {error1} {error2} }.
      //-----------------------------------------------------------------
      static const std::string updateExcludedDirList( const CHAR* const dir_list );

      //-----------------------------------------------------------------
      /// \brief Update file extension list
      ///
      /// \param[in] ext_list
      ///     A list of file extensions 
      ///
      /// \return
      ///     Nothing
      ///
      /// \exception runtime_error 
      ///     ext_list is empty 
      //-----------------------------------------------------------------
      static void updateFileExtList( const CHAR* const ext_list );
   
      //-----------------------------------------------------------------
      /// \brief Add/update directory to/in the global hash map.
      ///
      /// \param[in] dir
      ///     Full path for the directory as it appears in MOUNT_PT list.
      ///
      /// \return
      ///     Sorted Tcl list of scanned subdirectories that contain
      ///     frame data, followed by error messages if any:
      ///     { subdirs } { {error1} {error2} }.   
      //-----------------------------------------------------------------
      static const std::string addEntry( const std::string& dir );

      //-----------------------------------------------------------------
      /// \brief Permanently remove directory from global hash map.
      ///      
      /// \param[in] dir
      ///     Full path for the directory to remove.
      /// \param[out] errors
      ///     A list of errors encountered when trying to remove \a dir.
      ///
      /// \return
      ///     A list of removed subdirectories with frame data.
      //-----------------------------------------------------------------
      static const std::list< std::string >
      removeEntry( const std::string& dir,
                   LogMessage& errors );         

      //-----------------------------------------------------------------
      /// \brief Destroy existing data hash.
      ///      
      /// \return
      ///     Nothing.
      //-----------------------------------------------------------------
      static void removeEntries( );

      static const std::string getHashNumbers( const std::string& ifo, 
                                               const std::string& type );

      //-----------------------------------------------------------------
      /// \brief Get entries matching search criteria.
      ///
      /// \param ifo_type_str
      ///     An "IFO-Type-" string for frame data to look up.
      /// \param query_start
      ///     Query start time.
      /// \param query_stop
      ///     Query stop time.
      ///
      /// \return
      ///     Space seperated string of matching entries.
      //-----------------------------------------------------------------
      static const std::string getFileNames( const char* ifo_type_str,
                                             const INT_4U query_start,
                                             const INT_4U query_stop);

      static const std::string getIntervalsList( const std::string& ifo, 
                                                 const std::string& type, 
                                                 const INT_4U query_start,
                                                 const INT_4U query_stop );

      static const std::list< std::string > getFileExtList();   
   
   
      //: Helpers
      static const std::string
      createExcludedMsgHeader( const excluded_dirs_container_type& dir_pattern );
   
   
      //: Writer
      static void write( CacheFile& stream );
   
      //: Reader
      static void read( CacheFile& stream );   
   
   
   private:
      //=================================================================
      // Types - Private
      //=================================================================

      typedef std::vector< std::string >        ordered_mount_point_type;

      //-----------------------------------------------------------------
      //-----------------------------------------------------------------
      typedef General::ReadWriteLock lock_type;
      typedef MountPointCache::Info::key_type key_type;
      typedef MountPointCache::Info::key_base_type key_base_type;
      typedef General::SharedPtr< MountPointCache > mount_pt_cache_type;

      class cache_container_base_type
         : private General::unordered_map< key_type,
                                          mount_pt_cache_type,
                                          General::hash_pointer< key_base_type > >
      {
      public:
         typedef General::SharedConstPtr< DirectoryInfo > dir_info_type;
         cache_container_base_type( );
         ~cache_container_base_type( );

         //--------------------------------------------------------------
         /// \brief Search for information within the cache
         ///
         /// \param[in] SearchParams
         ///     The search parameters describing the data to find.
         /// \param[out] Results
         ///     The data found.
         //--------------------------------------------------------------
         void Find( const Query& SearchParams,
                    QueryResults& Results );

         void Register( dir_info_type DirInfo );

         //--------------------------------------------------------------
         // \brief Replace one node with another.
         //--------------------------------------------------------------
         void Replace( dir_info_type DirInfoOld,
                       dir_info_type DirInfoNew );

         void Unregister( dir_info_type DirInfo );
      private:
         typedef General::ReadWriteLock lock_type;

         lock_type::lock_type        m_baton;
      };
      typedef
#if ENABLE_MOUNT_POINT_CACHE
      General::SharedPtr< cache_container_base_type >
#else
      General::SharedPtr< DirectoryInfo >
#endif
      cache_container_type;

      //-----------------------------------------------------------------
      /// \brief Mount point container
      //-----------------------------------------------------------------
      typedef General::unordered_map< std::string, cache_container_type >
      MountPtContainer;

      //-----------------------------------------------------------------
      /// \brief Iterator for mount point container
      //-----------------------------------------------------------------
      typedef MountPtContainer::iterator MountPtIter;      

      //-----------------------------------------------------------------
      /// \brief Constant iterator for mount point container
      //-----------------------------------------------------------------
      typedef MountPtContainer::const_iterator MountPtConstIter;         

      //=================================================================
      // Methods - Private
      //=================================================================
      //-----------------------------------------------------------------
      //-----------------------------------------------------------------
      static const std::string
      validateMountPtList( std::vector< std::string >& mount_pt_list );
   
      static const std::string formatOutput( std::list< std::string >& data,
                                             const std::string& errors,
                                             const std::string& header = "" );
   
      //-----------------------------------------------------------------
      //-----------------------------------------------------------------
      static void cleanup();
   
      //-----------------------------------------------------------------
      //-----------------------------------------------------------------
      static const std::vector< TimeHash::KnownIntervals >
      getMountPtHashes( const std::string& ifo_type,
                        const std::string& mount_pt_dir, 
                        const dir_info_type except_dir );

      //-----------------------------------------------------------------
      /// \brief Get set of known "IFO-Type-" frame data.
      ///
      /// \param ifo_type
      ///     An IFO and Type for frame data.
      /// \param mount_pt_dir
      ///     A MOUNT_PT directory.
      /// \param query_start
      ///     Query start time.
      /// \param query_stop
      ///     Query stop time.
      /// \param except_dir
      ///     Don't collect interval information for this directory.
      ///     Default is ZERO.
      ///   
      /// \return
      ///     A set of already identified time ranges for "ifo_type"
      ///     frame data.
      //-----------------------------------------------------------------
      static const std::vector< TimeHash::KnownIntervals >
      getMountPtHashes( const std::string& ifo_type,
                        const std::string& mount_pt_dir, 
                        const INT_4U query_start, const INT_4U query_stop,
                        const dir_info_type except_dir );


      static excluded_dirs_rw_type excluded_dirs( );

      static mount_point_rw_type mount_points( const char* File = __FILE__,
                                               int Line = __LINE__ );

      //=================================================================
      // Data - Private
      //=================================================================
      static bool initialized;
      //-----------------------------------------------------------------
      //-----------------------------------------------------------------
      static lock_type::lock_type        mLock;
   
      //-----------------------------------------------------------------
      /// \brief Lock to preserve mFileExtList integrity
      //-----------------------------------------------------------------
      static lock_type::lock_type        mLockFileExtList;
      //-----------------------------------------------------------------
      /// \brief Flag to indicate if data conflicts should be checked across 
      ///     all MOUNT_PT entries.
      //-----------------------------------------------------------------
      static bool mDataConflictsAcrossAllMntPts;
   
   
      //-----------------------------------------------------------------
      /// \brief Sorted list of explicit directory names that should
      ///     not be scanned.
      //-----------------------------------------------------------------
      static std::list< std::string > mExcludedDirList;
   
      //-----------------------------------------------------------------
      /// \brief Sorted list of file extensions
      //-----------------------------------------------------------------
      static std::list< std::string > mFileExtList;
   
   
      //-----------------------------------------------------------------
      /// \brief Hash of MOUNT_PT_DIR <---> IFOTypeHash of known
      ///     frame data in all subdirectories
      //-----------------------------------------------------------------
      static MountPtContainer mMountPtHash;

      static MutexLock::lock_type       m_scanned_mount_points_baton;
      static INT_4U                     m_scanned_mount_points;
      static MutexLock::lock_type       m_last_update_baton;
      static General::GPSTime           m_last_update;

      //-----------------------------------------------------------------
      /// \brief Container of excluded directories
      //-----------------------------------------------------------------
      static excluded_dirs_container_type       p_excluded_dirs_container;
      //-----------------------------------------------------------------
      /// \brief Lock to preserve container of excluded directories integrity
      //-----------------------------------------------------------------
      static excluded_dirs_rw_type::lock_type   p_excluded_dirs_container_baton;
      //-----------------------------------------------------------------
      /// \brief Mount Point List as it appears in diskcacheAPI resource file
      //-----------------------------------------------------------------
      static mount_point_container_type         p_mount_point_container;
      //-----------------------------------------------------------------
      /// \brief Preserve the integrety of the mount point information.
      //-----------------------------------------------------------------
      static mount_point_ro_type::lock_type     p_mount_point_container_baton;
   
   };

   inline MountPointHash::Conflict::
   Conflict( const std::string& Message )
      : std::runtime_error( Message )
   {
   }

   inline bool MountPointHash::
   CheckForDataConflictsAcrossAllMountPoints( )
   {
      return mDataConflictsAcrossAllMntPts;
   }

   inline const MountPointHash::excluded_dirs_ro_type MountPointHash::
   ExcludedDirs( )
   {
      return( excluded_dirs_ro_type( p_excluded_dirs_container_baton,
                                     RWLOCK_TIMEOUT,
                                     p_excluded_dirs_container,
                                     __FILE__, __LINE__ ) )
         ;
   }

   inline const MountPointHash::mount_point_ro_type MountPointHash::
   MountPoints( const char* File, int Line )
   {
      return( mount_point_ro_type( p_mount_point_container_baton,
                                   RWLOCK_TIMEOUT,
                                   p_mount_point_container,
                                   File, Line ) )
         ;
   }

}
#endif   
