/* -*- c-basic-offset: 4; -*- */

#include "genericAPI/config.h"

#include "registry.hh"
#include "util.hh"


namespace Registry {

#if HAVE_LDAS_PACKAGE_ILWD
//-----------------------------------------------------------------------------
//
//: Element Registry
//
ElementRegistry_type	elementRegistry;
#endif /* HAVE_LDAS_PACKAGE_ILWD */


//-----------------------------------------------------------------------------
//
//: Binary Registry
//
BinaryRegistry_type	binaryRegistry;


//-----------------------------------------------------------------------------
//
//: Server Registry
//
ServerRegistry_type	serverRegistry;


//-----------------------------------------------------------------------------
//
//: Active Socket Registry
//
// These are sockets which are created with the createDataSocket command.
//
ObjectRegistry< os_tcp_socket >            activeSocketRegistry;


//-----------------------------------------------------------------------------
//: Passive Socket Registry
//
// These are sockets which are created with the acceptDataSocket command
//
ObjectRegistry< os_tcp_socket >            passiveSocketRegistry;


#if HAVE_LDAS_PACKAGE_ILWD
//-----------------------------------------------------------------------------
//
//: Check an Element Pointer
//
// True if the element is valid.
//
//!param: LdasElement* element
//
//!return: bool - True if the element is valid
//
bool isElementValid( const ILwd::LdasElement* element ) throw()
{
    return ( elementRegistry.isRegistered( element ) ||
             elementRegistry.contentsSearch( element ) );
}
#endif /* HAVE_LDAS_PACKAGE_ILWD */


//-----------------------------------------------------------------------------
//
//: Check Binary Pointer
//
// True if the LdasBinary object is valid.
//
//!param: Ldasbinary* binary
//
//!return: bool - True if the object is valid
//
bool isBinaryValid( const LdasBinary* binary ) throw()
{
    return ( binaryRegistry.isRegistered( binary ) );
}


//-----------------------------------------------------------------------------
//
//: Check Server Pointer
//
// True if the server is valid.
//
//!param: os_tcp_connection_server* socket
//
//!return: bool - True if the object is valid
//
bool isServerValid( const os_tcp_connection_server* server ) throw()
{
    return ( serverRegistry.isRegistered( server ) );
}


//-----------------------------------------------------------------------------
//
//: Check Socket Pointer
//
// True if the socket is valid.
//
//!param: os_tcp_socket* socket
//
//!return: bool - True if the socket is valid
//
bool isSocketValid( const os_tcp_socket* socket ) throw()
{
    if ( !activeSocketRegistry.isRegistered( socket ) )
    {
        if ( passiveSocketRegistry.isRegistered( socket ) )
        {
            return true;
        }
        else
        {
            return false;
        }
    }
    else
    {
        return true;
    }
}

//-----------------------------------------------------------------------------

#if HAVE_LDAS_PACKAGE_ILWD
bool ElementRegistry_type::
contentsSearch( const ILwd::LdasElement* Element ) const
{
    General::ReadWriteLock m( m_lock,
			      General::ReadWriteLock::READ );

    bool found = false;
    
    for( const_iterator iter = begin(),
	     last = end( );
	 ( iter != last ) && ( !found );
	 ++iter )
    {
        if ( iter->first->getElementId() == ILwd::ID_ILWD )
        {
            found =
		::contentsSearch( dynamic_cast< ILwd::LdasContainer& >
				  ( *(iter->first) ),
				  Element );
        }
    }

    return found;
}
#endif /* HAVE_LDAS_PACKAGE_ILWD */

#if HAVE_LDAS_PACKAGE_ILWD
std::string ElementRegistry_type::
getElementList( ) const
{
    General::ReadWriteLock m( m_lock,
			      General::ReadWriteLock::READ );

    std::stringstream ss;

    ss << "{ ";
    for ( const_iterator iter = begin(),
	      last = end( );
	  iter != last;
	  ++iter )
    {
        ss << makeTclPointer( iter->first, "LdasElement" ) << " ";
    }
    ss << "}";
    
    return ss.str();
}
#endif /* HAVE_LDAS_PACKAGE_ILWD */

} // namespace Registry

//-----------------------------------------------------------------------------
#if HAVE_LDAS_PACKAGE_ILWD
template class ObjectRegistry< ILwd::LdasElement >;
#endif /* HAVE_LDAS_PACKAGE_ILWD */
template class ObjectRegistry< LdasBinary >;
template class ObjectRegistry< os_tcp_connection_server >;
template class ObjectRegistry< os_tcp_socket >;
#if HAVE_TID
template class ObjectRegistry< tid >;
#endif /* HAVE_TID */
