#include "framecpp/config.h"

#include <memory.h>

#include <cmath>

#include "framecpp/FrameH.hh"

#include "framec/FrFrame.h"

#include "FrameCInternal.hh"
#include "StreamInternal.hh"

using FrameC::Deposit;
using FrameC::Set;
using FrameC::Validate;
using FrameC::Withdraw;

namespace FrameC
{
  struct FrFrame {
    ::fr_frame_t			m_data;
    std::auto_ptr< ::FrameCPP::FrameH >	m_frameh;

    FrFrame( );

    FrFrame( const char* Name,
	     fr_frame_t0_t Start,
	     fr_frame_dt_t Dt,
	     fr_frame_frame_t FrameNumber );

    ~FrFrame( );

    inline
    void Read( ::fr_file_t* Stream,
	       ::fr_frame_frame_t Offset )
    {
      FrameCPP::IFrameStream*	fs( StreamAsInput( Stream ) );

      m_frameh.reset( fs->ReadFrameH( Offset, 0 ) );
      
      if ( m_frameh.get( ) == (FrameCPP::FrameH*)NULL )
      {
	std::ostringstream	msg;

	msg << "Failed to read frame " << Offset << " from the stream.";

	throw std::runtime_error( msg.str( ) );
      }

      if ( m_data.name )
      {
	free( m_data.name );
      }
      m_data.name = ::strdup( m_frameh->GetName( ).c_str( ) );
      m_data.run = m_frameh->GetRun( );
      m_data.frame = m_frameh->GetFrame( );
      m_data.dq = m_frameh->GetDataQuality( );
      m_data.t0.sec = m_frameh->GetGTime( ).GetSeconds( );
      m_data.t0.nan = m_frameh->GetGTime( ).GetNanoseconds( );
      m_data.dt = (fr_frame_dt_t)( std::ceil( m_frameh->GetDt( ) ) );
    };

    inline
    void Write( ::fr_file_t* Stream )
    {
      FrameCPP::OFrameStream*	fs( StreamAsOutput( Stream ) );

      //-------------------------------------------------
      // Sync the C++ object with the C structure
      //-------------------------------------------------
      /// \todo Need to be able to set the name of an FrameH 
      /// frame structure
      //-------------------------------------------------
#if WORKING
      m_frameh->SetName( m_data.name );
#endif /* WORKING */
      m_frameh->SetRun( m_data.run );
      //-------------------------------------------------
      /// \todo Need to be able to set the frame of an FrameH 
      /// frame structure
      //-------------------------------------------------
#if WORKING
      m_frameh->SetFrame( m_data.frame );
#endif /* WORKING */
      m_frameh->SetDataQuality( m_data.dq );
      m_frameh->SetGTime( FrameCPP::Version::GPSTime( m_data.t0.sec,
						      m_data.t0.nan ) );
      m_frameh->SetDt( m_data.dt );
      //-------------------------------------------------
      // Write it out to disk
      //-------------------------------------------------
      fs->Write( m_frameh.get( ) );
    };
  };

  template<>
  inline Pointers::pointer_type
  PointerType( FrFrame* Value )
  {
    return Pointers::FR_FRAME;
  }

  template<>
  const char* Pointers::
  Name< Pointers::FR_FRAME >( )
  {
    return "FrFrame";
  }

}

fr_frame_t *
FrFrameAlloc( FrameCError** Error,
	      const char *Name,
	      fr_frame_t0_t t0,
	      fr_frame_dt_t dt,
	      fr_frame_frame_t frame )
{
  Set( Error );

  std::auto_ptr< FrameC::FrFrame >	tmp;
  fr_frame_t*				retval = (fr_frame_t*)NULL;

  try
  {
    tmp.reset( new FrameC::FrFrame( Name, t0, dt, frame ) );
  }
  catch( const std::exception& Exception )
  {
    Set( Error, FRAMEC_ERRNO_FR_FRAME_ALLOC_ERROR, Exception.what( ) );
    tmp.reset( (FrameC::FrFrame*)NULL );
  }

  if ( tmp.get( ) )
  {
    Deposit( tmp.get( ) );
    retval = &(tmp->m_data);
    tmp.release( );
  }
  return retval;
}

int
FrFrameFree( FrameCError** Error,
	     fr_frame_t *frame )
{
  Set( Error );

  int	retval = true;
  FrameC::FrFrame*	ooframe( (FrameC::FrFrame*)frame );

  try
  {
    if ( ooframe )
    {
      Withdraw( ooframe );
    }
  }
  catch( const std::exception& Exception )
  {
    Set( Error, FRAMEC_ERRNO_FR_FRAME_FREE_ERROR, Exception.what( ) );
    retval = false;
  }

  return retval;
}

fr_frame_t *
FrFrameRead( FrameCError** Error,
	     fr_file_t *restrict stream,
	     fr_frame_offset_t pos )
{
  Set( Error );

  std::auto_ptr< FrameC::FrFrame >	frame;

  try
  {
    frame.reset( new FrameC::FrFrame );
    frame->Read( stream, pos );
  }
  catch( const std::exception& Exception )
  {
    Set( Error, FRAMEC_ERRNO_FR_FRAME_READ_ERROR, Exception.what( ) );
    frame.reset( NULL );
  }
  catch( ... )
  {
    Set( Error, FRAMEC_ERRNO_FR_FRAME_READ_ERROR, FrameC::UnknownError );
    frame.reset( NULL );
  }

  fr_frame_t* retval = (fr_frame_t*)NULL;

  if ( frame.get( ) )
  {
    Deposit( frame.get( ) );
    retval = &(frame->m_data);
    frame.release( );
  }

  return retval;
}

int
FrFrameWrite( FrameCError** Error,
	      fr_file_t *restrict stream,
	      fr_frame_t *restrict frame )
{
  Set( Error );

  int	retval = false;

  try
  {
    FrameC::FrFrame*
      ooframe( reinterpret_cast< FrameC::FrFrame* >( frame ) );

    Validate( ooframe );

    ooframe->Write( stream );
    retval = true;
  }
  catch( const std::exception& Exception )
  {
    Set( Error, FRAMEC_ERRNO_FR_FRAME_WRITE_ERROR, Exception.what( ) );
  }
  catch( ... )
  {
    Set( Error, FRAMEC_ERRNO_FR_FRAME_WRITE_ERROR, FrameC::UnknownError );
  }

  return retval;
}

namespace FrameC
{
  FrFrame::
  FrFrame( )
  {
    m_data.name = (char*)NULL;
    m_data.run = -1;
    m_data.t0.sec = 0;
    m_data.t0.nan = 0;
    m_data.dt = 0;
    m_data.frame = 0;
  }

  FrFrame::
  FrFrame( const char* Name,
	   fr_frame_t0_t Start,
	   fr_frame_dt_t Dt,
	   fr_frame_frame_t FrameNumber )
  {
    m_data.name = ::strdup( Name );
    m_data.run = -1;
    m_data.t0 = Start;
    m_data.dt = Dt;
    m_data.frame = FrameNumber;

    General::GPSTime start( Start.sec, Start.nan );
    m_frameh.reset( new FrameCPP::FrameH( std::string( Name ),
					  0, /* run */
					  FrameNumber,
					  start,
					  start.GetLeapSeconds( ),
					  Dt ) );
  }

  FrFrame::
  ~FrFrame( )
  {
    free( m_data.name );
  }
}
