#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "framec/FrameC.h"
#include "framec/FrFrame.h"
#include "framec/Stream.h"

/*---------------------------------------------------------------*/

enum {
  PASS = 0,
  FAIL = 1
};

FrameCError*	ErrorCode;
int 		Status = PASS;

/*---------------------------------------------------------------*/

static const char*
pass_fail( int Check )
{
  if ( ! Check )
  {
    Status = FAIL;
  }
  return ( ( Check ) ? "PASS" : "FAIL" );
}

static void
log_test_result( const char* Routine,
		 int PassFail,
		 const char* Message )
{
  fprintf( stderr,
	   "%s: %s: %s\n",
	   pass_fail( PassFail ), Routine, Message );
}

/*---------------------------------------------------------------*/
void
ErrorCheck( const char* FunctionName, framec_errno_type EC )
{
  if ( ErrorCode )
  {
    const int s = ( ErrorCode->s_errno == EC );

    fprintf( stderr,
	     "%s: %s:error code: %d =?= %d [%s]\n",
	     pass_fail( s ), FunctionName, EC, ErrorCode->s_errno, ErrorCode->s_message );
  }
  else
  {
    const int s = ( FRAMEC_ERRNO_NO_ERROR == EC );

    fprintf( stderr,
	     "%s: %s:error code: %d =?= %d\n",
	     pass_fail( s ), FunctionName, EC, FRAMEC_ERRNO_NO_ERROR );
  }
}

static fr_frame_t*
generate_frame( int Offset, const char* Routine )
{
  fr_frame_t0_t 	t0 = { 600000000, 0 };
  fr_frame_dt_t		dt = 16.0;
  fr_frame_frame_t	fn = 1;
  fr_frame_t*		frameh = (fr_frame_t*)NULL;

  frameh = FrFrameAlloc( &ErrorCode, "LIGO", t0, dt, fn );
  if ( Routine ) { ErrorCheck( Routine, FRAMEC_ERRNO_NO_ERROR ); }

  return frameh;
}

static inline int
frameh_compare( const fr_frame_t* LHS,
		const fr_frame_t* RHS )
{
  int	retval = 0;

  if ( ( LHS != ( fr_frame_t*)NULL )
       && ( RHS != ( fr_frame_t*)NULL )
       && ( strcmp( LHS->name, RHS->name ) == 0 )
       && ( LHS->run == RHS->run )
       && ( LHS->frame == RHS->frame )
       && ( LHS->dq == RHS->dq )
       && ( LHS->t0.sec == RHS->t0.sec )
       && ( LHS->t0.nan == RHS->t0.nan )
       && ( LHS->dt == RHS->dt ) )
  {
    retval = 1;
  }
  return retval;
}

/*---------------------------------------------------------------*/
void
NoInputFile( )
{
  static const char* routine = "NoInputFile";
  /*
   * This test makes sure that an error is generated if a
   * request to open a non-existing file is made.
   */
  fr_file_t*	ifs;

  ifs = FrFileOpen( &ErrorCode, "junk.gwf", FRAMEC_FILE_MODE_INPUT );

  ErrorCheck( routine, FRAMEC_ERRNO_FRAME_OPEN_ERROR );
}

/*---------------------------------------------------------------*/
void
CloseUnknownFile( )
{
  static const char* routine = "CloseUnknownFile";
  /*
   * This test makes sure that an error is generated if a
   * request to open a non-existing file is made.
   */
  fr_file_t*	ofs;

  ofs = (fr_file_t*)( 3690 );
  FrFileClose( &ErrorCode, ofs );

  ErrorCheck( routine, FRAMEC_ERRNO_FRAME_CLOSE_ERROR );
}

/*---------------------------------------------------------------*/
void
DoubleClose( )
{
  static const char* routine = "DoubleClose";
  /*
   * This test makes sure that an error is generated if a
   * request to open a non-existing file is made.
   */
  fr_file_t*	ofs;

  ofs = FrFileOpen( &ErrorCode, "ojunk.gwf", FRAMEC_FILE_MODE_OUTPUT );
  ErrorCheck( routine, FRAMEC_ERRNO_NO_ERROR );

  FrFileClose( &ErrorCode, ofs );
  ErrorCheck( routine, FRAMEC_ERRNO_NO_ERROR );

  FrFileClose( &ErrorCode, ofs );
  ErrorCheck( routine, FRAMEC_ERRNO_FRAME_CLOSE_ERROR );

}
/*---------------------------------------------------------------*/
void
FrFrameCreation( )
{
  static const char* routine = "FrFrameCreation";

  fr_frame_t0_t 	t0 = { 600000000, 0 };
  fr_frame_dt_t		dt = 16.0;
  fr_frame_frame_t	fn = 1;
  fr_frame_t*		f = (fr_frame_t*)NULL;

  f = FrFrameAlloc( &ErrorCode, "LIGO", t0, dt, fn );
  ErrorCheck( routine, FRAMEC_ERRNO_NO_ERROR );

  FrFrameFree( &ErrorCode, f );
  ErrorCheck( routine, FRAMEC_ERRNO_NO_ERROR );
  
  FrFrameFree( &ErrorCode, f );
  ErrorCheck( routine, FRAMEC_ERRNO_FR_FRAME_FREE_ERROR );
}
/*---------------------------------------------------------------*/
void
StreamIO( )
{
  static const char* routine = "StreamIO";
  static const char* filename = "StreamIO.gwf";

  fr_file_t*		ofs = (fr_file_t*)NULL;
  int			test_result = 1;

  /*-------------------------------------------------------------
   * Generate a file to read
   *-------------------------------------------------------------*/
  ofs = FrFileOpen( &ErrorCode, filename, FRAMEC_FILE_MODE_OUTPUT );
  ErrorCheck( routine, FRAMEC_ERRNO_NO_ERROR );

  if ( ofs )
  {
    fr_frame_t*		frame0out = (fr_frame_t*)NULL;
    fr_frame_t*		frame0in = (fr_frame_t*)NULL;
    fr_file_t*		ifs = (fr_file_t*)NULL;

    /*-----------------------------------------------------------
     * Write the frame data to the file
     *-----------------------------------------------------------*/
    frame0out = generate_frame( 0, routine );
    FrFrameWrite( &ErrorCode, ofs, frame0out );
    ErrorCheck( routine, FRAMEC_ERRNO_NO_ERROR );

    if ( ofs ) { FrFileClose( &ErrorCode, ofs ); }
    /*-----------------------------------------------------------
     * Read the file
     *-----------------------------------------------------------*/

    ifs = FrFileOpen( &ErrorCode, filename, FRAMEC_FILE_MODE_INPUT );
    ErrorCheck( routine, FRAMEC_ERRNO_NO_ERROR );

    frame0in = FrFrameRead( &ErrorCode, ifs, 0 );
    ErrorCheck( routine, FRAMEC_ERRNO_NO_ERROR );
    
    test_result = frameh_compare( frame0out, frame0in );
    log_test_result( routine,
		     test_result, "Frame written is equivelant to the frame read." );

    /*-----------------------------------------------------------
     * Reclaim resources
     *-----------------------------------------------------------*/
    if ( frame0in ) { FrFrameFree( &ErrorCode, frame0in ); }
    if ( frame0out ) { FrFrameFree( &ErrorCode, frame0out ); }
    if ( ifs ) { FrFileClose( &ErrorCode, ifs ); }
  }

  /* unlink( filename ); */
}
/*---------------------------------------------------------------*/

/*---------------------------------------------------------------
 * This is main where the tests are actually executed.
 *---------------------------------------------------------------*/
int
main( int ArgC, char** ArgV )
{
  FrameCErrorInit( &ErrorCode );

  printf( "FrameC library version %d\n",
	  FrameCLibraryMinorVersion( ) );

  /*
   * Test that error conditions are properly detected.
   */
  /*--------------------------------------------------------------------*/
  NoInputFile( );
  CloseUnknownFile( );
  DoubleClose( );
  /*--------------------------------------------------------------------*/
  FrFrameCreation( );
  /*
   * Test Reading/Writing of a frame file
   */
  StreamIO( );
  return( Status );
}
